import torch
import torch.nn as nn


class conv_block(nn.Module):
    def __init__(self, ch_in, ch_out):
        super(conv_block, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(ch_in, ch_out, kernel_size=3, stride=1, padding=1, bias=True),
            nn.BatchNorm2d(ch_out),
            nn.ReLU(inplace=True),
            nn.Conv2d(ch_out, ch_out, kernel_size=3, stride=1, padding=1, bias=True),
            nn.BatchNorm2d(ch_out),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        x = self.conv(x)
        return x


class up_conv(nn.Module):
    def __init__(self, ch_in, ch_out):
        super(up_conv, self).__init__()
        self.up = nn.Sequential(
            nn.Upsample(scale_factor=2),
            nn.Conv2d(ch_in, ch_out, kernel_size=3, stride=1, padding=1, bias=True),
            nn.BatchNorm2d(ch_out),
            nn.ReLU(inplace=True),
        )

    def forward(self, x):
        x = self.up(x)
        return x


class RNNCell(nn.Module):
    def __init__(self, input_size, hidden_size):
        super(RNNCell, self).__init__()
        self.input_size = input_size
        self.hidden_size = hidden_size
        self.in_gate = nn.Conv2d(input_size + hidden_size, hidden_size, 3, 1, 1)
        self.forget_gate = nn.Conv2d(input_size + hidden_size, hidden_size, 3, 1, 1)
        self.out_gate = nn.Conv2d(input_size + hidden_size, hidden_size, 3, 1, 1)
        self.cell_gate = nn.Conv2d(input_size + hidden_size, hidden_size, 3, 1, 1)

    def forward(self, input, h_state, c_state):

        conc_inputs = torch.cat((input, h_state), 1)

        in_gate = self.in_gate(conc_inputs)
        forget_gate = self.forget_gate(conc_inputs)
        out_gate = self.out_gate(conc_inputs)
        cell_gate = self.cell_gate(conc_inputs)

        in_gate = torch.sigmoid(in_gate)
        forget_gate = torch.sigmoid(forget_gate)
        out_gate = torch.sigmoid(out_gate)
        cell_gate = torch.tanh(cell_gate)

        c_state = (forget_gate * c_state) + (in_gate * cell_gate)
        h_state = out_gate * torch.tanh(c_state)

        return h_state, c_state


class set_values(nn.Module):
    def __init__(self, hidden_size, height, width):
        super(set_values, self).__init__()
        self.hidden_size = int(hidden_size)
        self.height = int(height)
        self.width = int(width)
        self.dropout = nn.Dropout(0.7)
        self.RCell = RNNCell(self.hidden_size, self.hidden_size)

    def forward(self, seq, xinp):
        """[summary]

        Args:
            seq ([type]): A set of nn.module functions such as nn.Conv2d.
            xinp ([type]): [description]

        Returns:
            [type]: [description]
        """
        xout = torch.zeros(xinp.size()[0], xinp.size()[1], self.hidden_size, self.height, self.width).to(xinp.device)

        # h_state = torch.zeros(xinp[0].shape[0], self.hidden_size, self.height, self.width).to(xinp.device)
        # c_state = torch.zeros(xinp[0].shape[0], self.hidden_size, self.height, self.width).to(xinp.device)
        h_state = torch.zeros(xinp.shape[0], self.hidden_size, self.height, self.width).to(xinp.device)
        c_state = torch.zeros(xinp.shape[0], self.hidden_size, self.height, self.width).to(xinp.device)

        for t in range(xinp.shape[1]):
            input_t = seq(xinp[:, t])
            xout[:, t] = input_t
            h_state, c_state = self.RCell(input_t, h_state, c_state)

        return self.dropout(h_state), xout


class UNetLSTM(nn.Module):
    def __init__(self, in_channels, out_channels, patch_size):
        super(UNetLSTM, self).__init__()

        self.patch_size = patch_size
        self.Maxpool = nn.MaxPool2d(kernel_size=2, stride=2)

        self.Conv1 = conv_block(ch_in=in_channels, ch_out=16)
        self.set1 = set_values(16, self.patch_size, self.patch_size)

        self.Conv2 = conv_block(ch_in=16, ch_out=32)
        self.set2 = set_values(32, self.patch_size / 2, self.patch_size / 2)

        self.Conv3 = conv_block(ch_in=32, ch_out=64)
        self.set3 = set_values(64, self.patch_size / 4, self.patch_size / 4)

        self.Conv4 = conv_block(ch_in=64, ch_out=128)
        self.set4 = set_values(128, self.patch_size / 8, self.patch_size / 8)

        self.Conv5 = conv_block(ch_in=128, ch_out=256)
        self.set5 = set_values(256, self.patch_size / 16, self.patch_size / 16)

        self.Up5 = up_conv(ch_in=256, ch_out=128)
        self.Up_conv5 = conv_block(ch_in=256, ch_out=128)

        self.Up4 = up_conv(ch_in=128, ch_out=64)
        self.Up_conv4 = conv_block(ch_in=128, ch_out=64)

        self.Up3 = up_conv(ch_in=64, ch_out=32)
        self.Up_conv3 = conv_block(ch_in=64, ch_out=32)

        self.Up2 = up_conv(ch_in=32, ch_out=16)
        self.Up_conv2 = conv_block(ch_in=32, ch_out=16)

        self.Conv_1x1 = nn.Conv2d(16, out_channels, kernel_size=1, stride=1, padding=0)

    def encoder(self, x):
        x1, xout = self.set1(self.Conv1, x)

        x2, xout = self.set2(nn.Sequential(self.Maxpool, self.Conv2), xout)

        x3, xout = self.set3(nn.Sequential(self.Maxpool, self.Conv3), xout)

        x4, xout = self.set4(nn.Sequential(self.Maxpool, self.Conv4), xout)

        x5, xout = self.set5(nn.Sequential(self.Maxpool, self.Conv5), xout)

        return x1, x2, x3, x4, x5

    def forward(self, input):
        input = input.permute(1, 0, 2, 3, 4)
        # encoding path
        x1, x2, x3, x4, x5 = self.encoder(input)

        # decoding + concat path
        d5 = self.Up5(x5)
        d5 = torch.cat((d5, x4), dim=1)
        d5 = self.Up_conv5(d5)

        d4 = self.Up4(d5)
        d4 = torch.cat((d4, x3), dim=1)
        d4 = self.Up_conv4(d4)

        d3 = self.Up3(d4)
        d3 = torch.cat((d3, x2), dim=1)
        d3 = self.Up_conv3(d3)

        d2 = self.Up2(d3)
        d2 = torch.cat((d2, x1), dim=1)
        d2 = self.Up_conv2(d2)

        d1 = self.Conv_1x1(d2)
        d1 = torch.sigmoid(d1).squeeze(1)

        return d1


class UNetLSTMEncoder(nn.Module):
    def __init__(self, in_channels, patch_size):
        super(UNetLSTMEncoder, self).__init__()

        self.patch_size = patch_size
        self.Maxpool = nn.MaxPool2d(kernel_size=2, stride=2)

        self.Conv1 = conv_block(ch_in=in_channels, ch_out=16)
        self.set1 = set_values(16, self.patch_size, self.patch_size)

        self.Conv2 = conv_block(ch_in=16, ch_out=32)
        self.set2 = set_values(32, self.patch_size / 2, self.patch_size / 2)

        self.Conv3 = conv_block(ch_in=32, ch_out=64)
        self.set3 = set_values(64, self.patch_size / 4, self.patch_size / 4)

        self.Conv4 = conv_block(ch_in=64, ch_out=128)
        self.set4 = set_values(128, self.patch_size / 8, self.patch_size / 8)

        self.Conv5 = conv_block(ch_in=128, ch_out=256)
        self.set5 = set_values(256, self.patch_size / 16, self.patch_size / 16)

    def forward(self, x):
        """Compute forward pass of Encoder.

        Args:
            x [torch.Tensor]: A tensor of shape [batch_size, n_frames, n_channels, height, width].

        """

        x1, xout = self.set1(self.Conv1, x)

        x2, xout = self.set2(nn.Sequential(self.Maxpool, self.Conv2), xout)

        x3, xout = self.set3(nn.Sequential(self.Maxpool, self.Conv3), xout)

        x4, xout = self.set4(nn.Sequential(self.Maxpool, self.Conv4), xout)

        x5, xout = self.set5(nn.Sequential(self.Maxpool, self.Conv5), xout)

        return x1, x2, x3, x4, x5


class UNetLSTMDecoder(nn.Module):
    def __init__(self, out_channels):
        super(UNetLSTMDecoder, self).__init__()

        self.Up5 = up_conv(ch_in=256, ch_out=128)
        self.Up_conv5 = conv_block(ch_in=256, ch_out=128)

        self.Up4 = up_conv(ch_in=128, ch_out=64)
        self.Up_conv4 = conv_block(ch_in=128, ch_out=64)

        self.Up3 = up_conv(ch_in=64, ch_out=32)
        self.Up_conv3 = conv_block(ch_in=64, ch_out=32)

        self.Up2 = up_conv(ch_in=32, ch_out=16)
        self.Up_conv2 = conv_block(ch_in=32, ch_out=16)

        self.Conv_1x1 = nn.Conv2d(16, out_channels, kernel_size=1, stride=1, padding=0)

    def forward(self, feats):
        x1, x2, x3, x4, x5 = feats

        # decoding + concat path
        d5 = self.Up5(x5)
        d5 = torch.cat((d5, x4), dim=1)
        d5 = self.Up_conv5(d5)

        d4 = self.Up4(d5)
        d4 = torch.cat((d4, x3), dim=1)
        d4 = self.Up_conv4(d4)

        d3 = self.Up3(d4)
        d3 = torch.cat((d3, x2), dim=1)
        d3 = self.Up_conv3(d3)

        d2 = self.Up2(d3)
        d2 = torch.cat((d2, x1), dim=1)
        d2 = self.Up_conv2(d2)

        d1 = self.Conv_1x1(d2)
        d1 = torch.sigmoid(d1).squeeze(1)

        return d1



"""
Ported bits of netharn

import liberator
lib = liberator.Liberator()


import netharn as nh
lib.add_dynamic(nh.api.Initializer)

lib.expand(['netharn'])
print(lib.current_sourcecode())

lib.add_dynamic(nh.initializers.KaimingNormal)
print(lib.current_sourcecode())


import liberator
lib = liberator.Liberator()
lib.add_dynamic(nh.util.number_of_parameters)
print(lib.current_sourcecode())

"""
import torch
import ubelt as ub
import numpy as np  # NOQA


def _update_defaults(config, kw):
    config = dict(config)
    for k, v in kw.items():
        if k not in config:
            config[k] = v
    return config


class Initializer(object):
    """
    Base class for all netharn initializers
    """

    def __call__(self, model, *args, **kwargs):
        return self.forward(model, *args, **kwargs)

    def forward(self, model):
        """
        Abstract function that does the initailization
        """
        raise NotImplementedError('implement me')

    def history(self):
        """
        Initializer methods have histories which are short for algorithms and
        can be quite long for pretrained models
        """
        return None

    def get_initkw(self):
        """
        Initializer methods have histories which are short for algorithms and
        can be quite long for pretrained models
        """
        initkw = self.__dict__.copy()
        # info = {}
        # info['__name__'] = self.__class__.__name__
        # info['__module__'] = self.__class__.__module__
        # info['__initkw__'] = initkw
        return initkw

    @staticmethod
    def coerce(config={}, **kw):
        """
        Accepts 'init', 'pretrained', 'pretrained_fpath', 'leftover', and
        'noli'.

        Args:
            config (dict | str): coercable configuration dictionary.
                if config is a string it is taken as the value for "init".

        Returns:
            Tuple[Initializer, dict]: initializer_ = initializer_cls, kw

        Examples:
            >>> print(ub.urepr(Initializer.coerce({'init': 'noop'})))
            >>> config = {
            ...     'init': 'pretrained',
            ...     'pretrained_fpath': '/fit/nice/untitled'
            ... }
            >>> print(ub.urepr(Initializer.coerce(config)))
            >>> print(ub.urepr(Initializer.coerce({'init': 'kaiming_normal'})))
        """
        if isinstance(config, str):
            config = {
                'init': config,
            }

        config = _update_defaults(config, kw)

        pretrained_fpath = config.get('pretrained_fpath', config.get('pretrained', None))
        init = config.get('initializer', config.get('init', None))

        # Allow init to specify a pretrained fpath
        if isinstance(init, str) and pretrained_fpath is None:
            from os.path import exists
            pretrained_cand = ub.expandpath(init)
            if exists(pretrained_cand):
                pretrained_fpath = pretrained_cand

        config['init'] = init
        config['pretrained_fpath'] = pretrained_fpath
        config['pretrained'] = pretrained_fpath

        if pretrained_fpath is not None:
            config['init'] = 'pretrained'

        import os
        init_verbose = int(os.environ.get('WATCH_INIT_VERBOSE', 4))

        # ---
        initializer_ = None
        if config['init'].lower() in ['kaiming_normal']:
            initializer_ = (KaimingNormal, {
                # initialization params should depend on your choice of
                # nonlinearity in your model. See the Kaiming Paper for details.
                'param': 1e-2 if config.get('noli', 'relu') == 'leaky_relu' else 0,
            })
        elif config['init'] == 'noop':
            initializer_ = (NoOp, {})
        elif config['init'] == 'pretrained':
            from torch_liberator.initializer import Pretrained
            initializer_ = (Pretrained, {
                'fpath': ub.expandpath(config['pretrained_fpath']),
                'leftover': kw.get('leftover', None),
                'mangle': kw.get('mangle', False),
                'association': kw.get('association', None),
                'verbose': init_verbose,
            })
        elif config['init'] == 'cls':
            # Indicate that the model will initialize itself
            # We have to trust that the user does the right thing here.
            pass
        else:
            raise KeyError('Unknown coercable init: {!r}'.format(config['init']))
        return initializer_


class NoOp(Initializer):
    """
    An initializer that does nothing, which is useful when you have initialized
    the weights yourself.

    Example:
        >>> # xdoctest: +SKIP
        >>> import copy
        >>> self = NoOp()
        >>> model = toynet.ToyNet2d()
        >>> old_state = sum(v.sum() for v in model.state_dict().values())
        >>> self(model)
        >>> new_state = sum(v.sum() for v in model.state_dict().values())
        >>> assert old_state == new_state
        >>> assert self.history() is None
    """

    def forward(self, model):
        return


class Orthogonal(Initializer):
    """
    Same as Orthogonal, but uses pytorch implementation

    Example:
        >>> # xdoctest: +SKIP
        >>> self = Orthogonal()
        >>> model = toynet.ToyNet2d()
        >>> try:
        >>>     self(model)
        >>> except RuntimeError:
        >>>     import pytest
        >>>     pytest.skip('geqrf: Lapack probably not availble')
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """

    def __init__(self, gain=1):
        self.gain = gain

    def forward(self, model):
        try:
            func = torch.nn.init.orthogonal_
        except AttributeError:
            func = torch.nn.init.orthogonal

        apply_initializer(model, func, self.__dict__)


class KaimingUniform(Initializer):
    """
    Same as HeUniform, but uses pytorch implementation

    Example:
        >>> # xdoctest: +SKIP
        >>> from netharn.models import toynet
        >>> self = KaimingUniform()
        >>> model = toynet.ToyNet2d()
        >>> self(model)
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """

    def __init__(self, param=0, mode='fan_in'):
        self.a = param
        self.mode = mode

    def forward(self, model):
        try:
            func = torch.nn.init.kaiming_uniform_
        except AttributeError:
            func = torch.nn.init.kaiming_uniform
        apply_initializer(model, func, self.__dict__)


class KaimingNormal(Initializer):
    """
    Same as HeNormal, but uses pytorch implementation

    Example:
        >>> # xdoctest: +SKIP
        >>> from netharn.models import toynet
        >>> self = KaimingNormal()
        >>> model = toynet.ToyNet2d()
        >>> self(model)
        >>> layer = torch.nn.modules.Conv2d(3, 3, 3)
        >>> self(layer)
    """

    def __init__(self, param=0, mode='fan_in'):
        self.a = param
        self.mode = mode

    def forward(self, model):
        try:
            func = torch.nn.init.kaiming_normal_
        except AttributeError:
            func = torch.nn.init.kaiming_normal
        apply_initializer(model, func, self.__dict__)


def apply_initializer(input, func, funckw):
    """
    Recursively initializes the input using a torch.nn.init function.

    If the input is a model, then only known layer types are initialized.

    Args:
        input (Tensor | Module): can be a model, layer, or tensor
        func (callable): initialization function
        funckw (dict):

    Example:
        >>> from torch import nn
        >>> import torch
        >>> class DummyNet(nn.Module):
        >>>     def __init__(self, n_channels=1, n_classes=10):
        >>>         super(DummyNet, self).__init__()
        >>>         self.conv = nn.Conv2d(n_channels, 10, kernel_size=5)
        >>>         self.norm = nn.BatchNorm2d(10)
        >>>         self.param = torch.nn.Parameter(torch.rand(3))
        >>> self = DummyNet()
        >>> func = nn.init.kaiming_normal_
        >>> apply_initializer(self, func, {})
        >>> func = nn.init.constant_
        >>> apply_initializer(self, func, {'val': 42})
        >>> assert np.all(self.conv.weight.detach().numpy() == 42)
        >>> assert np.all(self.conv.bias.detach().numpy() == 0), 'bias is always init to zero'
        >>> assert np.all(self.norm.bias.detach().numpy() == 0), 'bias is always init to zero'
        >>> assert np.all(self.norm.weight.detach().numpy() == 1)
        >>> assert np.all(self.norm.running_mean.detach().numpy() == 0.0)
        >>> assert np.all(self.norm.running_var.detach().numpy() == 1.0)
    """
    if getattr(input, 'bias', None) is not None:
        # print('zero input bias')
        # zero all biases
        input.bias.data.zero_()

    if isinstance(input, (torch.Tensor)):
        # assert False, ('input is tensor? does this make sense?')
        # print('input is tensor')
        func(input, **funckw)
        # data = input
    elif isinstance(input, (torch.nn.modules.conv._ConvNd)):
        # print('input is convnd')
        func(input.weight, **funckw)
    # elif isinstance(input, (torch.nn.modules.linear.Linear)):
    #     func(input.weight, **funckw)
    elif isinstance(input, torch.nn.modules.batchnorm._BatchNorm):
        # Use default batch norm
        input.reset_parameters()
    # elif isinstance(input, torch.nn.modules.Linear):
    #     input.reset_parameters()
    elif hasattr(input, 'reset_parameters'):
        # print('unknown input type fallback on reset_params')
        input.reset_parameters()
    else:
        # input is a torch module
        model = input
        # print('recurse input')
        layers = list(trainable_layers(model))
        # print('layers = {!r}'.format(layers))
        for item in layers:
            apply_initializer(item, func, funckw)


def trainable_layers(model, names=False):
    """
    Returns all layers containing trainable parameters

    Notes:
        It may be better to simply use model.named_parameters() instead in most
        situation. This is useful when you need the classes that contains the
        parameters instead of the parameters themselves.

    Example:
        >>> import torchvision
        >>> model = torchvision.models.AlexNet()
        >>> list(trainable_layers(model, names=True))
    """
    if names:
        stack = [('', '', model)]
        while stack:
            prefix, basename, item = stack.pop()
            name = '.'.join([p for p in [prefix, basename] if p])
            if isinstance(item, torch.nn.modules.conv._ConvNd):
                yield name, item
            elif isinstance(item, torch.nn.modules.batchnorm._BatchNorm):
                yield name, item
            elif hasattr(item, 'reset_parameters'):
                yield name, item

            child_prefix = name
            for child_basename, child_item in list(item.named_children())[::-1]:
                stack.append((child_prefix, child_basename, child_item))
    else:
        queue = [model]
        while queue:
            item = queue.pop(0)
            # TODO: need to put all trainable layer types here
            # (I think this is just everything with reset_parameters)
            if isinstance(item, torch.nn.modules.conv._ConvNd):
                yield item
            elif isinstance(item, torch.nn.modules.batchnorm._BatchNorm):
                yield item
            elif hasattr(item, 'reset_parameters'):
                yield item
            # if isinstance(input, torch.nn.modules.Linear):
            #     yield item
            # if isinstance(input, torch.nn.modules.Bilinear):
            #     yield item
            # if isinstance(input, torch.nn.modules.Embedding):
            #     yield item
            # if isinstance(input, torch.nn.modules.EmbeddingBag):
            #     yield item
            for child in item.children():
                queue.append(child)


def number_of_parameters(model, trainable=True):
    """
    Returns number of trainable parameters in a torch module

    Example:
        >>> from watch.utils.util_netharn import *  # NOQA
        >>> model = torch.nn.Conv1d(2, 3, 5)
        >>> number_of_parameters(model)
        33
    """
    if trainable:
        model_parameters = filter(lambda p: p.requires_grad, model.parameters())
    else:
        model_parameters = model.parameters()
    n_params = sum([np.prod(p.size()) for p in model_parameters])
    return n_params



# flake8: noqa

import os
import comet_ml
import torch
import datetime
import random
import kwcoco
from torch import nn
import kwimage
import kwarray
import ndsampler
import watch.tasks.rutgers_material_seg.utils.eval_utils as eval_utils
import numpy as np
import torch.nn.functional as F
from tqdm import tqdm  # NOQA
import ubelt as ub
import pathlib
import watch.tasks.rutgers_material_seg.utils.utils as utils
from watch.tasks.rutgers_material_seg.models import build_model
from watch.tasks.rutgers_material_seg.datasets.iarpa_contrastive_dataset import SequenceDataset
from torchvision import transforms
from skimage.filters import threshold_otsu as otsu
import matplotlib.pyplot as plt
import gc
import watch.tasks.rutgers_material_seg.utils.visualization as visualization

current_path = os.getcwd().split("/")


class Evaluator(object):
    def __init__(self,
                 model: object,
                 eval_loader: torch.utils.data.DataLoader,
                 config,
                 device='cuda'):
        """Evaluator class

        Args:
            model (object): trained or untrained model
            eval_loader (torch.utils.data.DataLader): loader with evaluation data
            optimizer (object): optimizer to train with
            scheduler (object): scheduler to train with
        """

        self.model = model
        self.eval_loader = eval_loader
        self.device = device
        self.config = config
        self.num_classes = self.config['data']['num_classes']
        self.max_label = self.num_classes
        self.inference_all_crops_params = [tuple([i, j, config['evaluation']['inference_window'], config['evaluation']['inference_window']]) for i in range(0, config['data']['image_size']) for j in range(0, config['data']['image_size'])]
        self.cmap = visualization.n_distinguishable_colors(nlabels=self.max_label,
                                                    first_color_black=True, last_color_black=True,
                                                    bg_alpha=config['visualization']['bg_alpha'],
                                                    fg_alpha=config['visualization']['fg_alpha'])

    def high_confidence_filter(self, features: torch.Tensor, cutoff_top: float = 0.75,
                               cutoff_low: float = 0.0, eps: float = 1e-8) -> torch.Tensor:
        """Select high confidence regions to select as predictions

        Args:
            features (torch.Tensor): initial mask
            cutoff_top (float, optional): cutoff of the object. Defaults to 0.75.
            cutoff_low (float, optional): low cutoff. Defaults to 0.2.
            eps (float, optional): small number. Defaults to 1e-8.

        Returns:
            torch.Tensor: pseudo mask generated
        """
        bs, c, h, w = features.size()
        features = features.view(bs, c, -1)

        # for each class extract the max confidence
        features_max, _ = features.max(-1, keepdim=True)
        # features_max[:, c-1:] *= 0.8
        # features_max[:, :c-1] *= cutoff_top
        features_max *= cutoff_top
        # features_max *= cutoff_top

        # if the top score is too low, ignore it
        lowest = torch.Tensor([cutoff_low]).type_as(features_max)
        features_max = features_max.max(lowest)

        filtered_features = (features > features_max).type_as(features)
        filtered_features = filtered_features.view(bs, c, h, w) * features.view(bs, c, h, w)
        return filtered_features

    def eval(self, cometml_experiemnt) -> tuple:
        """evaluate a single epoch

        Args:

        Returns:
            None
        """
        vw_disagreement, histogram_distance, l1_dist, l2_dist = [], [], [], []
        topk_pre_histogram_distance, topk_pre_l1_dist, topk_pre_l2_dist = [], [], []
        topk_post_histogram_distance, topk_post_l1_dist, topk_post_l2_dist = [], [], []
        targets = []

        self.model.eval()

        with torch.no_grad():
            # Prog = ub.ProgIter
            Prog = tqdm
            pbar = Prog(enumerate(self.eval_loader), total=len(self.eval_loader), desc='predict rutgers')
            for batch_index, batch in pbar:
                outputs = batch
                images, mask = outputs['inputs']['im'].data[0], batch['label']['class_masks'].data[0]

                mask = torch.stack(mask)
                mask = mask.long().squeeze(1)

                bs, c, t, h, w = images.shape

                assert images.shape[2] == 2, 'only handles 2 frames'

                image1 = images[:, :, 0, :, :]
                image2 = images[:, :, 1, :, :]
                mask1 = mask[:, 0, :, :]  # NOQA
                mask2 = mask[:, 1, :, :]  # NOQA

                images = images.to(self.device)
                image1 = image1.to(self.device)
                image2 = image2.to(self.device)
                mask = mask.to(self.device)

                image1 = utils.stad_image(image1)
                image2 = utils.stad_image(image2)

                output1, features1 = self.model(image1)  # [B,22,150,150]
                output2, features2 = self.model(image2)

                masks1 = F.softmax(output1, dim=1)  # .detach()
                masks2 = F.softmax(output2, dim=1)  # .detach()
                masks1 = self.high_confidence_filter(masks1,
                                                     cutoff_top=self.config['high_confidence_threshold']['val_cutoff'],
                                                     cutoff_low=self.config['high_confidence_threshold']['val_low_cutoff'])
                masks2 = self.high_confidence_filter(masks2,
                                                     cutoff_top=self.config['high_confidence_threshold']['val_cutoff'],
                                                     cutoff_low=self.config['high_confidence_threshold']['val_low_cutoff'])

                pred1 = masks1.max(1)[1].cpu().detach()  # .numpy()
                pred2 = masks2.max(1)[1].cpu().detach()  # .numpy()

                vw_disagreement_pred = (pred1 != pred2).type(torch.uint8)

                inference_otsu_coeff = 1.5
                hist_inference_otsu_coeff = 0.95
                pad_amount = (self.config['evaluation']['inference_window'] - 1) // 2
                topk = 13

                padded_output1 = F.pad(input=output1, pad=(pad_amount, pad_amount, pad_amount, pad_amount), mode='replicate')
                padded_output2 = F.pad(input=output2, pad=(pad_amount, pad_amount, pad_amount, pad_amount), mode='replicate')
                patched_padded_output1 = torch.stack([transforms.functional.crop(padded_output1, *params) for params in self.inference_all_crops_params], dim=1)  # .flatten(-3,-1)
                patched_padded_output2 = torch.stack([transforms.functional.crop(padded_output2, *params) for params in self.inference_all_crops_params], dim=1)  # .flatten(-3,-1)

                padded_mask1 = F.pad(input=masks1, pad=(pad_amount, pad_amount, pad_amount, pad_amount), mode='replicate')
                padded_mask2 = F.pad(input=masks2, pad=(pad_amount, pad_amount, pad_amount, pad_amount), mode='replicate')
                patched_padded_mask1 = torch.stack([transforms.functional.crop(padded_mask1, *params) for params in self.inference_all_crops_params], dim=1)  # .flatten(-3,-1)
                patched_padded_mask2 = torch.stack([transforms.functional.crop(padded_mask2, *params) for params in self.inference_all_crops_params], dim=1)  # .flatten(-3,-1)

                patched_padded_output1_distributions = patched_padded_output1.flatten(-2, -1)  # .sum(axis=3) #[bs, n_patches, k]
                patched_padded_output2_distributions = patched_padded_output2.flatten(-2, -1)  # .sum(axis=3) #[bs, n_patches, k]
                patched_padded_mask1_distributions = patched_padded_mask1.flatten(-2, -1)  # .sum(axis=3) #[bs, n_patches, k]
                patched_padded_mask2_distributions = patched_padded_mask2.flatten(-2, -1)  # .sum(axis=3) #[bs, n_patches, k]

                topk_patched_output1_pre_distributions, largest_elements_pre_inds = torch.topk(patched_padded_output1_distributions, k=topk, sorted=False, dim=3)
                topk_patched_output2_pre_distributions = torch.gather(patched_padded_output2_distributions, dim=3, index=largest_elements_pre_inds)

                patched_padded_output1_distributions = patched_padded_output1_distributions.sum(axis=3)  # [bs, n_patches, k]
                patched_padded_output2_distributions = patched_padded_output2_distributions.sum(axis=3)  # [bs, n_patches, k]
                topk_patched_padded_output1_pre_distributions = topk_patched_output1_pre_distributions.sum(axis=3)  # [bs, n_patches, k]
                topk_patched_padded_output2_pre_distributions = topk_patched_output2_pre_distributions.sum(axis=3)  # [bs, n_patches, k]

                patched_padded_mask1_distributions = patched_padded_mask1_distributions.sum(axis=3)  # [bs, n_patches, k]
                patched_padded_mask2_distributions = patched_padded_mask2_distributions.sum(axis=3)  # [bs, n_patches, k]

                topk_patched_output1_post_distributions, largest_elements_post_inds = torch.topk(patched_padded_output1_distributions, k=topk, sorted=False, dim=2)
                topk_patched_output2_post_distributions = torch.gather(patched_padded_output2_distributions, dim=2, index=largest_elements_post_inds)

                normalized_topk_patched_output1_pre_distributions = (topk_patched_output1_pre_distributions - topk_patched_output1_pre_distributions.min(dim=2, keepdim=True)[0]) / (topk_patched_output1_pre_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output1_pre_distributions.min(dim=2, keepdim=True)[0])
                normalized_topk_patched_output2_pre_distributions = (topk_patched_output2_pre_distributions - topk_patched_output2_pre_distributions.min(dim=2, keepdim=True)[0]) / (topk_patched_output2_pre_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output2_pre_distributions.min(dim=2, keepdim=True)[0])
                normalized_topk_patched_output1_post_distributions = (topk_patched_output1_post_distributions - topk_patched_output1_post_distributions.min(dim=2, keepdim=True)[0]) / (topk_patched_output1_post_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output1_post_distributions.min(dim=2, keepdim=True)[0])
                normalized_topk_patched_output2_post_distributions = (topk_patched_output2_post_distributions - topk_patched_output2_post_distributions.min(dim=2, keepdim=True)[0]) / (topk_patched_output2_post_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output2_post_distributions.min(dim=2, keepdim=True)[0])

                # l1 region-wise inference raw features
                l1_patched_diff_change_features = torch.abs((patched_padded_output1_distributions - patched_padded_output2_distributions).sum(axis=2)).view(bs, h, w)
                l1_dist_change_feats_pred = torch.zeros_like(l1_patched_diff_change_features)
                l1_inference_otsu_threshold = inference_otsu_coeff * otsu(l1_patched_diff_change_features.cpu().detach().numpy(), nbins=256)
                l1_dist_change_feats_pred[l1_patched_diff_change_features > l1_inference_otsu_threshold] = 1
                l1_dist_change_feats_pred = l1_dist_change_feats_pred.cpu().detach().type(torch.uint8)

                # l1 region-wise inference pre topk
                l1_patched_diff_change_pre_topk = torch.abs((topk_patched_padded_output1_pre_distributions - topk_patched_padded_output2_pre_distributions).sum(axis=2)).view(bs, h, w)
                l1_dist_change_feats_pred_pre_topk = torch.zeros_like(l1_patched_diff_change_pre_topk)
                l1_inference_otsu_threshold = inference_otsu_coeff * otsu(l1_patched_diff_change_pre_topk.cpu().detach().numpy(), nbins=256)
                l1_dist_change_feats_pred_pre_topk[l1_patched_diff_change_pre_topk > l1_inference_otsu_threshold] = 1
                l1_dist_change_feats_pred_pre_topk = l1_dist_change_feats_pred_pre_topk.cpu().detach().type(torch.uint8)

                # l1 region-wise inference post topk
                l1_patched_diff_change_post_topk = torch.abs((topk_patched_output1_post_distributions - topk_patched_output2_post_distributions).sum(axis=2)).view(bs, h, w)
                l1_dist_change_feats_pred_post_topk = torch.zeros_like(l1_patched_diff_change_post_topk)
                l1_inference_otsu_threshold = inference_otsu_coeff * otsu(l1_patched_diff_change_post_topk.cpu().detach().numpy(), nbins=256)
                l1_dist_change_feats_pred_post_topk[l1_patched_diff_change_post_topk > l1_inference_otsu_threshold] = 1
                l1_dist_change_feats_pred_post_topk = l1_dist_change_feats_pred_post_topk.cpu().detach().type(torch.uint8)

                # l2 region-wise inference raw features
                l2_patched_diff_change_features = torch.sqrt(torch.pow(patched_padded_output1_distributions - patched_padded_output2_distributions, 2).sum(axis=2)).view(bs, h, w)
                l2_dist_change_feats_pred = torch.zeros_like(l2_patched_diff_change_features)
                l2_inference_otsu_threshold = inference_otsu_coeff * otsu(l2_patched_diff_change_features.cpu().detach().numpy(), nbins=256)
                l2_dist_change_feats_pred[l2_patched_diff_change_features > l2_inference_otsu_threshold] = 1
                l2_dist_change_feats_pred = l2_dist_change_feats_pred.cpu().detach().type(torch.uint8)

                # l2 region-wise inference pre topk
                l2_patched_diff_change_pre_topk = torch.sqrt(torch.pow(topk_patched_padded_output1_pre_distributions - topk_patched_padded_output2_pre_distributions, 2).sum(axis=2)).view(bs, h, w)
                l2_dist_change_feats_pred_pre_topk = torch.zeros_like(l2_patched_diff_change_pre_topk)
                l2_inference_otsu_threshold = inference_otsu_coeff * otsu(l2_patched_diff_change_pre_topk.cpu().detach().numpy(), nbins=256)
                l2_dist_change_feats_pred_pre_topk[l2_patched_diff_change_pre_topk > l2_inference_otsu_threshold] = 1
                l2_dist_change_feats_pred_pre_topk = l2_dist_change_feats_pred_pre_topk.cpu().detach().type(torch.uint8)

                # l2 region-wise inference  post topk
                l2_patched_diff_change_post_topk = torch.sqrt(torch.pow(patched_padded_output1_distributions - patched_padded_output2_distributions, 2).sum(axis=2)).view(bs, h, w)
                l2_dist_change_feats_pred_post_topk = torch.zeros_like(l2_patched_diff_change_post_topk)
                l2_inference_otsu_threshold = inference_otsu_coeff * otsu(l2_patched_diff_change_post_topk.cpu().detach().numpy(), nbins=256)
                l2_dist_change_feats_pred_post_topk[l2_patched_diff_change_post_topk > l2_inference_otsu_threshold] = 1
                l2_dist_change_feats_pred_post_topk = l2_dist_change_feats_pred_post_topk.cpu().detach().type(torch.uint8)

                # histogram intersection raw features
                # normalized_patched_padded_output1_distributions = (patched_padded_output1_distributions - patched_padded_output1_distributions.min(dim=2, keepdim=True)[0])/(patched_padded_output1_distributions.max(dim=2, keepdim=True)[0] - patched_padded_output1_distributions.min(dim=2, keepdim=True)[0])
                # normalized_patched_padded_output2_distributions = (patched_padded_output2_distributions - patched_padded_output2_distributions.min(dim=2, keepdim=True)[0])/(patched_padded_output2_distributions.max(dim=2, keepdim=True)[0] - patched_padded_output2_distributions.min(dim=2, keepdim=True)[0])
                minima = torch.minimum(patched_padded_mask1_distributions, patched_padded_mask2_distributions)
                # minima = torch.minimum(masks1, masks2)
                histograms_intersection_features = torch.true_divide(minima.sum(axis=2), patched_padded_mask2_distributions.sum(axis=2)).view(bs, h, w)

                histc_int_change_feats_pred = torch.zeros_like(histograms_intersection_features)
                histc_int_inference_otsu_threshold = hist_inference_otsu_coeff * otsu(histograms_intersection_features.cpu().detach().numpy(), nbins=256)
                histc_int_change_feats_pred[histograms_intersection_features < histc_int_inference_otsu_threshold] = 1
                histc_int_change_feats_pred = histc_int_change_feats_pred.cpu().detach().type(torch.uint8)

                # histogram intersection pre topk
                # normalized_topk_patched_output1_pre_distributions = (topk_patched_output1_pre_distributions - topk_patched_output1_pre_distributions.min(dim=2, keepdim=True)[0])/(topk_patched_output1_pre_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output1_pre_distributions.min(dim=2, keepdim=True)[0])
                # normalized_topk_patched_output2_pre_distributions = (topk_patched_output2_pre_distributions - topk_patched_output2_pre_distributions.min(dim=2, keepdim=True)[0])/(topk_patched_output2_pre_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output2_pre_distributions.min(dim=2, keepdim=True)[0])
                # minima_pre_topk = torch.minimum(topk_patched_output1_pre_distributions, topk_patched_output2_pre_distributions)
                # print(minima_pre_topk.shape)
                # print(topk_patched_output2_pre_distributions.shape)
                # histograms_intersection_pre_topk = torch.true_divide(minima_pre_topk.sum(axis=2), topk_patched_output2_pre_distributions.sum(axis=2)).view(bs,h,w)

                # histc_int_change_feats_pred_pre_topk = torch.zeros_like(histograms_intersection_pre_topk)
                # histc_int_inference_otsu_threshold = hist_inference_otsu_coeff*otsu(histograms_intersection_pre_topk.cpu().detach().numpy(), nbins=256)
                # histc_int_change_feats_pred_pre_topk[histograms_intersection_pre_topk < histc_int_inference_otsu_threshold] = 1
                # histc_int_change_feats_pred_pre_topk = histc_int_change_feats_pred_pre_topk.cpu().detach().type(torch.uint8)

                # histogram intersection post topk
                normalized_topk_patched_output1_post_distributions = (topk_patched_output1_post_distributions - topk_patched_output1_post_distributions.min(dim=2, keepdim=True)[0]) / (topk_patched_output1_post_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output1_post_distributions.min(dim=2, keepdim=True)[0])
                normalized_topk_patched_output2_post_distributions = (topk_patched_output2_post_distributions - topk_patched_output2_post_distributions.min(dim=2, keepdim=True)[0]) / (topk_patched_output2_post_distributions.max(dim=2, keepdim=True)[0] - topk_patched_output2_post_distributions.min(dim=2, keepdim=True)[0])
                minima_post_topk = torch.minimum(normalized_topk_patched_output1_post_distributions, normalized_topk_patched_output2_post_distributions)
                histograms_intersection_post_topk = torch.true_divide(minima_post_topk.sum(axis=2), topk_patched_output2_post_distributions.sum(axis=2)).view(bs, h, w)
                histc_int_change_feats_pred_post_topk = torch.zeros_like(histograms_intersection_post_topk)
                histc_int_inference_otsu_threshold = hist_inference_otsu_coeff * otsu(histograms_intersection_post_topk.cpu().detach().numpy(), nbins=256)
                histc_int_change_feats_pred_post_topk[histograms_intersection_post_topk < histc_int_inference_otsu_threshold] = 1
                histc_int_change_feats_pred_post_topk = histc_int_change_feats_pred_post_topk.cpu().detach().type(torch.uint8)

                # print(histc_int_change_feats_pred.shape)

                vw_disagreement.append(vw_disagreement_pred)
                histogram_distance.append(histc_int_change_feats_pred)
                l1_dist.append(l1_dist_change_feats_pred)
                l2_dist.append(l2_dist_change_feats_pred)
                topk_pre_l1_dist.append(l1_dist_change_feats_pred_pre_topk)
                topk_pre_l2_dist.append(l2_dist_change_feats_pred_pre_topk)
                topk_post_l1_dist.append(l1_dist_change_feats_pred_post_topk)
                topk_post_l2_dist.append(l2_dist_change_feats_pred_post_topk)

                # topk_pre_histogram_distance.append(histc_int_change_feats_pred_pre_topk)
                topk_post_histogram_distance.append(histc_int_change_feats_pred_post_topk)

                mask1[mask1 == -1] = 0
                targets.append(mask1.cpu())

                if self.config['visualization']['val_visualizer']:

                    batch_index_to_show = 0
                    figure = plt.figure(figsize=(
                        self.config['visualization']['fig_size'], self.config['visualization']['fig_size']))
                    ax1 = figure.add_subplot(3, 4, 1)
                    ax2 = figure.add_subplot(3, 4, 2)
                    ax3 = figure.add_subplot(3, 4, 3)
                    ax4 = figure.add_subplot(3, 4, 4)
                    ax5 = figure.add_subplot(3, 4, 5)
                    ax6 = figure.add_subplot(3, 4, 6)
                    ax7 = figure.add_subplot(3, 4, 7)
                    ax8 = figure.add_subplot(3, 4, 8)
                    ax9 = figure.add_subplot(3, 4, 9)
                    ax10 = figure.add_subplot(3, 4, 10)
                    ax11 = figure.add_subplot(3, 4, 11)
                    ax12 = figure.add_subplot(3, 4, 12)

                    cmap_gradients = plt.cm.get_cmap('jet')

                    image_show1 = np.transpose(image1.cpu().detach().numpy()[batch_index_to_show, :, :, :], (1, 2, 0))[:, :, :3]
                    image_show1 = np.flip(image_show1, axis=2)

                    image_show2 = np.transpose(image2.cpu().detach().numpy()[batch_index_to_show, :, :, :], (1, 2, 0))[:, :, :3]
                    image_show2 = np.flip(image_show2, axis=2)

                    image_show1 = (image_show1 - image_show1.min()) / (image_show1.max() - image_show1.min())
                    image_show2 = (image_show2 - image_show2.min()) / (image_show2.max() - image_show2.min())

                    gt_mask_show1 = mask1.cpu().detach()[batch_index_to_show, :, :].numpy().squeeze()

                    l1_dist_change_feats_pred_show = l1_dist_change_feats_pred.numpy()[batch_index_to_show, :, :]
                    l2_dist_change_feats_pred_show = l2_dist_change_feats_pred.numpy()[batch_index_to_show, :, :]
                    histc_int_change_feats_pred_show = histc_int_change_feats_pred.numpy()[batch_index_to_show, :, :]
                    vw_disagreement_pred_show = vw_disagreement_pred.numpy()[batch_index_to_show, :, :]

                    l1_patched_diff_change_features_show = l1_patched_diff_change_features.cpu().detach().numpy()[batch_index_to_show, :, :]
                    l2_patched_diff_change_features_show = l2_patched_diff_change_features.cpu().detach().numpy()[batch_index_to_show, :, :]
                    histograms_intersection_show = histograms_intersection_features.cpu().detach().numpy()[batch_index_to_show, :, :]

                    l1_patched_diff_change_features_show = (l1_patched_diff_change_features_show - l1_patched_diff_change_features_show.min()) / (l1_patched_diff_change_features_show.max() - l1_patched_diff_change_features_show.min())
                    l2_patched_diff_change_features_show = (l2_patched_diff_change_features_show - l2_patched_diff_change_features_show.min()) / (l2_patched_diff_change_features_show.max() - l2_patched_diff_change_features_show.min())
                    histograms_intersection_show = (histograms_intersection_show - histograms_intersection_show.min()) / (histograms_intersection_show.max() - histograms_intersection_show.min())

                    pred1_show = masks1.max(1)[1].cpu().detach().numpy()[batch_index_to_show, :, :]
                    pred2_show = masks2.max(1)[1].cpu().detach().numpy()[batch_index_to_show, :, :]

                    l1_fp_tp_fn_prediction_mask = gt_mask_show1 + (2 * l1_dist_change_feats_pred_show)
                    l2_fp_tp_fn_prediction_mask = gt_mask_show1 + (2 * l2_dist_change_feats_pred_show)
                    histc_fp_tp_fn_prediction_mask = gt_mask_show1 + (2 * histc_int_change_feats_pred_show)
                    vw_dis_fp_tp_fn_prediction_mask = gt_mask_show1 + (2 * vw_disagreement_pred_show)

                    pred1_show[pred1_show == -1] = 0
                    pred2_show[pred2_show == -1] = 0
                    gt_mask_show_no_bg1 = np.ma.masked_where(gt_mask_show1 == 0, gt_mask_show1)
                    # gt_mask_show_no_bg2 = np.ma.masked_where(gt_mask_show2==0,gt_mask_show2)
                    # logits_show_no_bg = np.ma.masked_where(logits_show==0,logits_show)

                    classes_in_gt = np.unique(gt_mask_show1)
                    ax1.imshow(image_show1)

                    ax2.imshow(image_show2)

                    ax3.imshow(image_show1)
                    ax3.imshow(gt_mask_show1, cmap=self.cmap, vmin=0, vmax=self.max_label)

                    ax4.imshow(image_show1)
                    ax4.imshow(pred1_show, cmap=self.cmap, vmin=0, vmax=self.max_label)

                    ax5.imshow(l1_patched_diff_change_features_show)

                    ax6.imshow(l2_patched_diff_change_features_show)

                    ax7.imshow(histograms_intersection_show)

                    ax8.imshow(image_show2)
                    ax8.imshow(pred2_show, cmap=self.cmap, vmin=0, vmax=self.max_label)

                    ax9.imshow(l1_fp_tp_fn_prediction_mask, cmap=self.cmap, vmin=0, vmax=self.max_label)

                    ax10.imshow(l2_fp_tp_fn_prediction_mask, cmap=self.cmap, vmin=0, vmax=self.max_label)

                    ax11.imshow(histc_fp_tp_fn_prediction_mask, cmap=self.cmap, vmin=0, vmax=self.max_label)

                    ax12.imshow(vw_dis_fp_tp_fn_prediction_mask, cmap=self.cmap, vmin=0, vmax=self.max_label)

                    if self.config['visualization']['titles']:
                        ax1.set_title(f"image_show1", fontsize=self.config['visualization']['font_size'])
                        ax2.set_title(f"image_show2", fontsize=self.config['visualization']['font_size'])
                        ax3.set_title(f"gt_mask_show1", fontsize=self.config['visualization']['font_size'])
                        ax4.set_title(f"pred1_show", fontsize=self.config['visualization']['font_size'])
                        ax5.set_title(f"l1_patched_diff_change_features_show", fontsize=self.config['visualization']['font_size'])
                        ax6.set_title(f"l2_patched_diff_change_features_show", fontsize=self.config['visualization']['font_size'])
                        ax7.set_title(f"histograms_intersection_show", fontsize=self.config['visualization']['font_size'])
                        ax8.set_title(f"pred2_show", fontsize=self.config['visualization']['font_size'])
                        ax9.set_title(f"l1_fp_tp_fn_prediction_mask", fontsize=self.config['visualization']['font_size'])
                        ax10.set_title(f"l2_fp_tp_fn_prediction_mask", fontsize=self.config['visualization']['font_size'])
                        ax11.set_title(f"histc_fp_tp_fn_prediction_mask", fontsize=self.config['visualization']['font_size'])
                        ax12.set_title(f"vw_dis_fp_tp_fn_prediction_mask", fontsize=self.config['visualization']['font_size'])
                        # figure.suptitle(
                        #     f"Epoch: {epoch+1}\nGT labels for classification: {classes_in_gt}, \nunique in change predictions: {np.unique(change_detection_show)}\nunique in predictions1: {np.unique(logits_show1)}", fontsize=config['visualization']['font_size'])

                    figure.tight_layout()
                    if self.config['visualization']['val_imshow']:
                        plt.show()

                    cometml_experiemnt.log_figure(figure_name=f"Validation", figure=figure)
                    figure.clear()
                    plt.cla()
                    plt.clf()
                    plt.close('all')
                    plt.close(figure)
                    gc.collect()

        l1_mean_iou, l1_precision, l1_recall = eval_utils.compute_jaccard(l1_dist, targets, num_classes=2)
        l2_mean_iou, l2_precision, l2_recall = eval_utils.compute_jaccard(l2_dist, targets, num_classes=2)

        l1_mean_iou_pre_topk, l1_precision_pre_topk, l1_recall_pre_topk = eval_utils.compute_jaccard(topk_pre_l1_dist, targets, num_classes=2)
        l2_mean_iou_pre_topk, l2_precision_pre_topk, l2_recall_pre_topk = eval_utils.compute_jaccard(topk_pre_l2_dist, targets, num_classes=2)
        l1_mean_iou_post_topk, l1_precision_post_topk, l1_recall_post_topk = eval_utils.compute_jaccard(topk_post_l1_dist, targets, num_classes=2)
        l2_mean_iou_post_topk, l2_precision_post_topk, l2_recall_post_topk = eval_utils.compute_jaccard(topk_post_l2_dist, targets, num_classes=2)

        vw_disagreement_mean_iou, vw_disagreement_precision, vw_disagreement_recall = eval_utils.compute_jaccard(vw_disagreement, targets, num_classes=2)
        hist_mean_iou, hist_precision, hist_recall = eval_utils.compute_jaccard(histogram_distance, targets, num_classes=2)

        # hist_mean_iou_pre_topk, hist_precision_pre_topk, hist_recall_pre_topk = eval_utils.compute_jaccard(topk_pre_histogram_distance, targets, num_classes=2)
        hist_mean_iou_post_topk, hist_precision_post_topk, hist_recall_post_topk = eval_utils.compute_jaccard(topk_post_histogram_distance, targets, num_classes=2)

        l1_precision = np.array(l1_precision)
        l1_recall = np.array(l1_recall)
        l1_f1 = 2 * (l1_precision * l1_recall) / (l1_precision + l1_recall)

        l2_precision = np.array(l2_precision)
        l2_recall = np.array(l2_recall)
        l2_f1 = 2 * (l2_precision * l2_recall) / (l2_precision + l2_recall)

        l1_precision_pre_topk = np.array(l1_precision_pre_topk)
        l1_recall_pre_topk = np.array(l1_recall_pre_topk)
        l1_f1_pre_topk = 2 * (l1_precision_pre_topk * l1_recall_pre_topk) / (l1_precision_pre_topk + l1_recall_pre_topk)

        l2_precision_pre_topk = np.array(l2_precision_pre_topk)
        l2_recall_pre_topk = np.array(l2_recall_pre_topk)
        l2_f1_pre_topk = 2 * (l2_precision_pre_topk * l2_recall_pre_topk) / (l2_precision_pre_topk + l2_recall_pre_topk)

        l1_precision_post_topk = np.array(l1_precision_post_topk)
        l1_recall_post_topk = np.array(l1_recall_post_topk)
        l1_f1_post_topk = 2 * (l1_precision_post_topk * l1_recall_post_topk) / (l1_precision_post_topk + l1_recall_post_topk)

        l2_precision_post_topk = np.array(l2_precision_post_topk)
        l2_recall_post_topk = np.array(l2_recall_post_topk)
        l2_f1_post_topk = 2 * (l2_precision_post_topk * l2_recall_post_topk) / (l2_precision_post_topk + l2_recall_post_topk)

        vw_disagreement_precision = np.array(vw_disagreement_precision)
        vw_disagreement_recall = np.array(vw_disagreement_recall)
        vw_disagreement_f1 = 2 * (vw_disagreement_precision * vw_disagreement_recall) / (vw_disagreement_precision + vw_disagreement_recall)

        hist_precision = np.array(hist_precision)
        hist_recall = np.array(hist_recall)
        hist_f1 = 2 * (hist_precision * hist_recall) / (hist_precision + hist_recall)

        # hist_precision_pre_topk = np.array(hist_precision_pre_topk)
        # hist_recall_pre_topk = np.array(hist_recall_pre_topk)
        # hist_f1_pre_topk = 2 * (hist_precision_pre_topk * hist_recall_pre_topk) / (hist_precision_pre_topk + hist_recall_pre_topk)

        hist_precision_post_topk = np.array(hist_precision_post_topk)
        hist_recall_post_topk = np.array(hist_recall_post_topk)
        hist_f1_post_topk = 2 * (hist_precision_post_topk * hist_recall_post_topk) / (hist_precision_post_topk + hist_recall_post_topk)

        print("\n")
        print({f"l1_recall class {str(x)}": l1_recall[x] for x in range(len(l1_recall))})
        print({f"l1_precision class {str(x)}": l1_precision[x] for x in range(len(l1_precision))})
        print({f"l1_f1 class {str(x)}": l1_f1[x] for x in range(len(l1_f1))})

        print("\n")
        print({f"l2_recall class {str(x)}": l2_recall[x] for x in range(len(l2_recall))})
        print({f"l2_precision class {str(x)}": l2_precision[x] for x in range(len(l2_precision))})
        print({f"l2_f2 class {str(x)}": l2_f1[x] for x in range(len(l2_f1))})

        print("\n")
        print({f"PRE TOPK l1_recall class {str(x)}": l1_recall_pre_topk[x] for x in range(len(l1_recall_pre_topk))})
        print({f"PRE TOPK l1_precision class {str(x)}": l1_precision_pre_topk[x] for x in range(len(l1_precision_pre_topk))})
        print({f"PRE TOPK l1_f1 class {str(x)}": l1_f1_pre_topk[x] for x in range(len(l1_f1_pre_topk))})

        print("\n")
        print({f"PRE TOPK l2_recall class {str(x)}": l2_recall_pre_topk[x] for x in range(len(l2_recall_pre_topk))})
        print({f"PRE TOPK l2_precision class {str(x)}": l2_precision_pre_topk[x] for x in range(len(l2_precision_pre_topk))})
        print({f"PRE TOPK l2_f2 class {str(x)}": l2_f1_pre_topk[x] for x in range(len(l2_f1_pre_topk))})

        print("\n")
        print({f"POST TOPK l1_recall class {str(x)}": l1_recall_post_topk[x] for x in range(len(l1_recall_post_topk))})
        print({f"POST TOPK l1_precision class {str(x)}": l1_precision_post_topk[x] for x in range(len(l1_precision_post_topk))})
        print({f"POST TOPK l1_f1 class {str(x)}": l1_f1_post_topk[x] for x in range(len(l1_f1_post_topk))})

        print("\n")
        print({f"POST TOPK l2_recall class {str(x)}": l2_recall_post_topk[x] for x in range(len(l2_recall_post_topk))})
        print({f"POST TOPK l2_precision class {str(x)}": l2_precision_post_topk[x] for x in range(len(l2_precision_post_topk))})
        print({f"POST TOPK l2_f2 class {str(x)}": l2_f1_post_topk[x] for x in range(len(l2_f1_post_topk))})

        print("\n")
        print({f"vw_disagreement_recall class {str(x)}": vw_disagreement_recall[x] for x in range(len(vw_disagreement_recall))})
        print({f"vw_disagreement_precision class {str(x)}": vw_disagreement_precision[x] for x in range(len(vw_disagreement_precision))})
        print({f"vw_disagreement_f1 class {str(x)}": vw_disagreement_f1[x] for x in range(len(vw_disagreement_f1))})

        print("\n")
        print({f"hist_recall class {str(x)}": hist_recall[x] for x in range(len(hist_recall))})
        print({f"hist_precision class {str(x)}": hist_precision[x] for x in range(len(hist_precision))})
        print({f"hist_f1 class {str(x)}": hist_f1[x] for x in range(len(hist_f1))})

        # print("\n")
        # print({f"PRE TOPK hist_recall class {str(x)}": hist_recall_pre_topk[x] for x in range(len(hist_recall_pre_topk))})
        # print({f"PRE TOPK hist_precision class {str(x)}": hist_precision_pre_topk[x] for x in range(len(hist_precision_pre_topk))})
        # print({f"PRE TOPK hist_f2 class {str(x)}": hist_f1_pre_topk[x] for x in range(len(hist_f1_pre_topk))})

        print("\n")
        print({f"POST TOPK hist_recall class {str(x)}": hist_recall_post_topk[x] for x in range(len(hist_recall_post_topk))})
        print({f"POST TOPK hist_precision class {str(x)}": hist_precision_post_topk[x] for x in range(len(hist_precision_post_topk))})
        print({f"POST TOPK hist_f1 class {str(x)}": hist_f1_post_topk[x] for x in range(len(hist_f1_post_topk))})
        return

    def forward(self, cometml_experiemnt) -> tuple:
        """forward pass for all epochs

        Args:
            cometml_experiment (object): comet ml experiment for logging
            world_size (int, optional): for distributed training. Defaults to 8.

        Returns:
            tuple: (train losses, validation losses, mIoU)
        """

        if self.config['procedures']['validate']:
            self.eval(cometml_experiemnt)
        return


def main(cmdline=True, **kwargs):
    """
    Ignore:
        # Hack in overrides because none of this is parameterized
        # state_dict = torch.load(checkpoint_fpath)
        checkpoint_fpath = ub.expandpath("$HOME/data/dvc-repos/smart_watch_dvc/models/rutgers/experiments_epoch_30_loss_0.05691597167379317_valmIoU_0.5694727912477856_time_2021-08-07-09:01:01.pth")
        cmdline = False
        kwargs = dict(
            default_config_key='iarpa',
            checkpoint_fpath=checkpoint_fpath,
            test_dataset=ub.expandpath("$HOME/data/dvc-repos/smart_watch_dvc/drop1-S2-L8-aligned/data.kwcoco.json"),
            pred_dataset='./test-pred/pred.kwcoco.json',
        )
    """
    project_root = '/'.join(current_path[:-1])
    # main_config_path = f"{os.getcwd()}/configs/main.yaml"
    main_config_path = f"{project_root}/configs/main.yaml"

    initial_config = utils.load_yaml_as_dict(main_config_path)
    # experiment_config_path = f"{os.getcwd()}/configs/{initial_config['dataset']}.yaml"
    experiment_config_path = f"{project_root}/configs/{initial_config['dataset']}.yaml"
    # config_path = utils.dictionary_contents(os.getcwd()+"/",types=["*.yaml"])[0]

    experiment_config = utils.config_parser(experiment_config_path, experiment_type="training")
    config = {**initial_config, **experiment_config}
    config['start_time'] = datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')

    project_name = f"{current_path[-3]}_{current_path[-1]}_{config['dataset']}"  # _{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M')}"
    experiment_name = f"attention_{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')}"
    experiment = comet_ml.Experiment(api_key=config['cometml']['api_key'],
                                     project_name=project_name,
                                     workspace=config['cometml']['workspace'],
                                     display_summary_level=0)
    experiment.set_name(experiment_name)

    torch.manual_seed(config['seed'])
    torch.cuda.manual_seed(config['seed'])
    np.random.seed(config['seed'])
    random.seed(config['seed'])
    torch.set_default_dtype(torch.float32)

    device_ids = list(range(torch.cuda.device_count()))
    config['device_ids'] = device_ids
    gpu_devices = ','.join([str(id) for id in device_ids])
    os.environ["CUDA_VISIBLE_DEVICES"] = gpu_devices
    device = torch.device('cuda')

    config['devices_used'] = gpu_devices
    experiment.log_asset_data(config)
    experiment.log_text(config)
    experiment.log_parameters(config)
    experiment.log_parameters(config['training'])
    experiment.log_parameters(config['evaluation'])
    experiment.log_parameters(config['visualization'])

    # print(config['data']['image_size'])
    coco_fpath = ub.expandpath(config['data'][config['location']]['train_coco_json'])
    dset = kwcoco.CocoDataset(coco_fpath)
    sampler = ndsampler.CocoSampler(dset)

    if config['training']['resume'] != False:
        base_path = '/'.join(config['training']['resume'].split('/')[:-1])
        pretrain_config_path = f"{base_path}/config.yaml"
        pretrain_config = utils.load_yaml_as_dict(pretrain_config_path)
        # print(config['training']['model_feats_channels'])
        # print(pretrain_config_path['training']['model_feats_channels'])
        config['data']['channels'] = pretrain_config['data']['channels']
        # if not config['training']['model_feats_channels'] == pretrain_config_path['training']['model_feats_channels']:
        #     print("the loaded model does not have the same number of features as configured in the experiment yaml file. Matching channel sizes to the loaded model instead.")
        # config['training']['model_feats_channels'] = pretrain_config_path['training']['model_feats_channels']
        config['data']['num_classes'] = pretrain_config['data']['num_classes']
        config['training']['model_feats_channels'] = pretrain_config['training']['model_feats_channels']

    window_dims = (config['data']['time_steps'], config['data']['image_size'], config['data']['image_size'])  # [t,h,w]
    input_dims = (config['data']['image_size'], config['data']['image_size'])

    # channels = 'B01|B02|B03|B04|B05|B06|B07|B08|B09|B10|B11|B12|B8A'
    channels = config['data']['channels']
    num_channels = len(channels.split('|'))
    config['training']['num_channels'] = num_channels
    # channels = 'red|green|blue'
    # channels = 'gray'
    dataset = SequenceDataset(sampler, window_dims, input_dims, channels)
    print(dataset.__len__())
    eval_dataloader = dataset.make_loader(batch_size=config['training']['batch_size'])

    model = build_model(model_name=config['training']['model_name'],
                        backbone=config['training']['backbone'],
                        pretrained=config['training']['pretrained'],
                        num_classes=config['data']['num_classes'],
                        num_groups=config['training']['gn_n_groups'],
                        weight_std=config['training']['weight_std'],
                        beta=config['training']['beta'],
                        num_channels=config['training']['num_channels'],
                        out_dim=config['training']['out_features_dim'],
                        feats=config['training']['model_feats_channels'])

    # model = SupConResNet(name=config['training']['backbone'])
    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("model has {} trainable parameters".format(num_params))
    model = nn.DataParallel(model)
    model.to(device)

    if config['training']['resume'] != False:

        if os.path.isfile(config['training']['resume']):
            checkpoint = torch.load(config['training']['resume'])
            model.load_state_dict(checkpoint['model'], strict=False)
            print(f"loaded model from {config['training']['resume']}")
        else:
            print("no checkpoint found at {}".format(config['training']['resume']))
            exit()

    evaler = Evaluator(model,
                      eval_dataloader,
                      config=config
                      )
    self = evaler  # NOQA
    evaler.forward(experiment)


if __name__ == "__main__":
    main()



# flake8: noqa
# from material_seg.datasets import build_dataset
import os
import sys
import comet_ml
import kwcoco
import ndsampler
import ubelt as ub
import watch
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import torch
import pdb
from sklearn.cluster import KMeans, SpectralClustering
from sklearn.manifold import TSNE
import torch.nn as nn
import torch.optim as optim
import datetime
import random

from watch.tasks.rutgers_material_seg.datasets.iarpa_dataset import SequenceDataset
from watch.tasks.rutgers_material_seg.models import build_model
import watch.tasks.rutgers_material_seg.utils.utils as utils


class Clusterer(object):
    def __init__(self, loader: torch.utils.data.DataLoader,
                 optimizer: object, scheduler: object,
                 clusterer: str = "kmeans", k: int = 20, **clusterer_kwargs) -> None:
        """clustering object

        Parameters
        ----------
        loader : torch.utils.data.DataLoader
            data to cluster
        optimizer : object
            optimizer used to train pre-trained model
        scheduler : object
            scheduler of pre-trained model
        clusterer : str, optional
            clustering algorithm, by default "kmeans"
        k : int, optional
            number of clusters, by default 20
        """
        self.models = {'kmeans': KMeans, 'spectralcluster': SpectralClustering}
        self.clustering_model = self.models[clusterer](n_clusters=k, **clusterer_kwargs)
        self.clustering_model_tsne = self.models[clusterer](n_clusters=k, **clusterer_kwargs)
        self.use_crf = config['evaluation']['use_crf']
        self.loader = loader
        self.optimizer = optimizer
        self.scheduler = scheduler

        self.cmap = visualization.rand_cmap(nlabels=config['data']['num_classes'] + 1, type='bright',
                                            first_color_black=True, last_color_black=True,
                                            bg_alpha=config['visualization']['bg_alpha'],
                                            fg_alpha=config['visualization']['fg_alpha'])

    def multitemporal_cluster(self):
        """Clustering algorithm considers features from multiple timesteps.
        """
        for batch in loader:
            # pdb.set_trace()
            image_data = batch['inputs']['im'].data[0]  # [b,c,t,h,w]
            b, c, t, h, w = image_data.shape
            mask_data = batch['label']['class_masks'].data[0]  # len(mask_data) = b
            mask_data = torch.stack(mask_data)
            outputs = torch.zeros((b, num_classes, t, h, w))
            for sub_batch in range(b):
                for timestep in range(t):
                    b_t_image_input = image_data[sub_batch, :, timestep, :, :].unsqueeze(0)
                    output = model(b_t_image_input)  # [1,num_classes + 1, h, w]
                    output.cpu().detach()
                    outputs[sub_batch, :, timestep, :, :] = output

            image_show = np.array(image_data).transpose(0, 2, 3, 4, 1) / 500  # visualize 0 indexed in batch
            # mask_show = np.array(mask_data) # [b,t,h,w]

            b, c, t, h, w = outputs.shape
            image_data = outputs.view(b, c * t, h * w).detach()
            image_data = torch.transpose(image_data, 1, 2)
            image_data = torch.flatten(image_data, start_dim=0, end_dim=1)

            out_feat_embed = TSNE(n_components=2).fit_transform(image_data)

            self.clustering_model.fit(image_data)
            self.clustering_model_tsne.fit(out_feat_embed)
            # cluster_centers = clustering_model.cluster_centers_
            cluster_labels = self.clustering_model.labels_
            y_kmeans = self.clustering_model_tsne.labels_
            prediction = cluster_labels.reshape(h, w)
            prediction_no_bg = np.ma.masked_where(prediction == 0, prediction)
            # print(f"image_data: {image_data.shape}, mask: {mask_data.shape}")
            # print(f"image min: {image_show.min()}, image max: {image_show.max()}")
            plt.scatter(out_feat_embed[:, 0], out_feat_embed[:, 1], c=y_kmeans, marker='.', cmap='tab20c')
            # plt.scatter(clustering_model_tsne.cluster_centers_[:, 0], clustering_model_tsne.cluster_centers_[:, 1], c='black', s=200, alpha=0.5);
            plt.show()

            figure = plt.figure(figsize=(15, 15))
            ax1 = figure.add_subplot(1, 5, 1)
            ax2 = figure.add_subplot(1, 5, 2)
            ax3 = figure.add_subplot(1, 5, 3)
            ax4 = figure.add_subplot(1, 5, 4)
            ax5 = figure.add_subplot(1, 5, 5)
            ax1.imshow(image_show[0, 0, :, :, :])
            ax2.imshow(image_show[0, 1, :, :, :])
            ax3.imshow(image_show[0, 2, :, :, :])
            ax4.imshow(image_show[0, 3, :, :, :])
            ax5.imshow(prediction, vmin=0, vmax=k, cmap='tab20c')
            plt.show()

            if visualize_images:
                mask_show = mask_show[0]  # [b,t,h,w]
                image_show = image_show[0]
                figure = plt.figure(figsize=(10, 10))
                axes = {}
                for i in range(1, 2 * t + 1):
                    axes[i] = figure.add_subplot(2, t, i)
                for key in axes.keys():
                    if key <= t:
                        axes[key].imshow(image_show[key - 1, :, :, :])
                    else:
                        axes[key].imshow(mask_show[key - t - 1, :, :], vmin=-1, vmax=7)
                figure.tight_layout()
                plt.show()


if __name__ == "__main__":

    project_root = "/home/native/projects/watch/watch/tasks/rutgers_material_seg/"
    # main_config_path = f"{os.getcwd()}/configs/main.yaml"
    # main_config_path = f"{os.getcwd()}/configs/main.yaml"
    main_config_path = f"{project_root}/configs/main.yaml"
    initial_config = utils.load_yaml_as_dict(main_config_path)
    experiment_config_path = f"{project_root}/configs/{initial_config['dataset']}.yaml"

    experiment_config = utils.config_parser(experiment_config_path, experiment_type="training")
    config = {**initial_config, **experiment_config}
    config['start_time'] = datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')

    project_name = f"{project_root[-3]}_{project_root[-1]}"  # _{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M')}"
    experiment_name = f"SMART_{datetime.datetime.today().strftime('%Y-%m-%d-%H:%M:%S')}"
    experiment = comet_ml.Experiment(api_key=config['cometml']['api_key'],
                                     project_name=project_name,
                                     workspace=config['cometml']['workspace'],
                                     display_summary_level=0)
    experiment.set_name(experiment_name)

    torch.manual_seed(config['seed'])
    torch.cuda.manual_seed(config['seed'])
    np.random.seed(config['seed'])
    random.seed(config['seed'])
    torch.set_default_dtype(torch.float32)

    # device_cpu = torch.device('cpu')
    # print(config['data']['image_size'])
    device_ids = list(range(torch.cuda.device_count()))
    config['device_ids'] = device_ids
    gpu_devices = ','.join([str(id) for id in device_ids])
    os.environ["CUDA_VISIBLE_DEVICES"] = gpu_devices
    device = torch.device('cuda')

    config['devices_used'] = gpu_devices
    experiment.log_asset_data(config)
    experiment.log_text(config)
    experiment.log_parameters(config)
    experiment.log_parameters(config['training'])
    experiment.log_parameters(config['evaluation'])
    experiment.log_parameters(config['visualization'])

    channels = 'r|g|b'
    number_of_timestamps, h, w = 4, 128, 128
    window_dims = (number_of_timestamps, h, w)  # [t,h,w]
    input_dims = (h, w)

    coco_fpath = ub.expandpath(config['data'][config['location']]['coco_json'])
    dset = kwcoco.CocoDataset(coco_fpath)

    sampler = ndsampler.CocoSampler(dset)
    # # channels = 'r|g|b|gray|wv1'
    dataset = SequenceDataset(sampler, window_dims, input_dims, channels)
    train_dataloader = dataset.make_loader(batch_size=config['training']['batch_size'])

    model = build_model(model_name=config['training']['model_name'],
                        backbone=config['training']['backbone'],
                        pretrained=config['training']['pretrained'],
                        num_classes=config['data']['num_classes'] + 1,
                        num_groups=config['training']['gn_n_groups'],
                        weight_std=config['training']['weight_std'],
                        beta=config['training']['beta'])

    model = build_model(model_name=config['training']['model_name'],
                        backbone=config['training']['backbone'],
                        pretrained=config['training']['pretrained'],
                        num_classes=config['data']['num_classes'] + 1,
                        num_groups=config['training']['gn_n_groups'],
                        weight_std=config['training']['weight_std'],
                        beta=config['training']['beta'])

    num_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    print("model has {} trainable parameters".format(num_params))
    model = nn.DataParallel(model)
    model.to(device)

    for m in model.modules():
        if isinstance(m, nn.BatchNorm2d):
            m.eval()
            m.weight.requires_grad = False
            m.bias.requires_grad = False

    optimizer = optim.SGD(model.parameters(),
                          lr=config['training']['learning_rate'],
                          momentum=config['training']['momentum'],
                          weight_decay=config['training']['weight_decay'])

    scheduler = optim.lr_scheduler.CosineAnnealingLR(optimizer, len(train_dataloader),
                                                     eta_min=config['training']['learning_rate'])

    if config['training']['resume'] != False:

        if os.path.isfile(config['training']['resume']):
            checkpoint = torch.load(config['training']['resume'])
            start_epoch = checkpoint['epoch']
            model.load_state_dict(checkpoint['model'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            scheduler.load_state_dict(checkpoint['scheduler'])
            print(f"loaded model from {config['training']['resume']}")
        else:
            print("no checkpoint found at {}".format(config['training']['resume']))
            exit()

    clusterer = Clusterer(model,
                      train_dataloader,
                      train_dataloader,
                      config['training']['epochs'],
                      optimizer,
                      scheduler,
                      test_loader=train_dataloader
                      )
    train_losses, val_losses, mean_ious_val = clusterer.multitemporal_cluster(experiment)



import ubelt as ub
import scriptconfig as scfg
import os


class CocoReformatChannels(scfg.Config):
    """
    Helper to remove channels from a coco file and reformat predictions from
    float32 to int16.
    """
    __default__ = {
        'src': scfg.Value(None, help='path to coco dataset to read and rewrite'),

        'quantize': scfg.Value(None, type=str, help='channels to quantize'),

        'remove': scfg.Value(None, type=str, help='channels to remove'),

        'workers': scfg.Value(0, help='number of background workers'),

        'nodata': scfg.Value(-9999, help='nodata value to use'),
    }


def main(cmdline=False, **kwargs):
    """
    Example:
        >>> from watch.cli.coco_reformat_channels import *  # NOQA
        >>> import kwcoco
        >>> import kwarray
        >>> import shutil
        >>> # Make a dataset we can modify inplace
        >>> orig_dset = kwcoco.CocoDataset.demo('vidshapes1-msi')
        >>> orig_dset.reroot(absolute=False)
        >>> orig_bundle = ub.Path(orig_dset.bundle_dpath)
        >>> new_bundle = ub.Path.appdir('kwcoco/tests/test_reformat_channels')
        >>> new_bundle.delete().ensuredir()
        >>> shutil.copytree(orig_bundle, new_bundle, dirs_exist_ok=True)
        >>> new_fpath = new_bundle / 'data.kwcoco.json'
        >>> orig_dset.dump(new_fpath)
        >>> new_dset = kwcoco.CocoDataset(new_fpath)
        >>> gid = 1
        >>> coco_img = new_dset.coco_image(gid)
        >>> rng = kwarray.ensure_rng(432)
        >>> imdata = rng.rand(128, 128)
        >>> coco_img.add_auxiliary_item('test_prediction.tif', channels='salient', imdata=imdata, imwrite=True)
        >>> new_dset.dump(new_dset.fpath)
        >>> # now reformat this new dataset
        >>> orig_pred = coco_img.imdelay('salient').finalize()
        >>> kwargs = {
        >>>     'src': new_dset.fpath,
        >>>     'quantize': 'salient',
        >>>     #'remove': 'B11',
        >>> }
        >>> cmdline = False
        >>> main(cmdline=False, **kwargs)
        >>> reformatted_dset = kwcoco.CocoDataset(new_fpath)
        >>> assert 'quantization' in reformatted_dset.imgs[1]['auxiliary'][-1]
        >>> new_coco_img = reformatted_dset.coco_image(gid)
        >>> import numpy as np
        >>> new_pred1 = np.nan_to_num(new_coco_img.imdelay('salient').finalize())
        >>> #assert np.allclose(new_pred1, new_pred2)
        >>> #new_pred2 = new_coco_img.imdelay('salient').finalize(dequantize=False)
        >>> #assert new_pred2.dtype.kind == 'i'
    """
    config = CocoReformatChannels.cli(data=kwargs, cmdline=cmdline)
    import kwcoco
    print('config = {}'.format(ub.urepr(dict(config), nl=1)))
    dset = kwcoco.CocoDataset.coerce(config['src'])

    to_quantize = kwcoco.ChannelSpec.coerce('' if config['quantize'] is None else config['quantize']).fuse()
    to_remove = kwcoco.ChannelSpec.coerce('' if config['remove'] is None else config['remove']).fuse()

    tasks = []
    for coco_img in dset.images().coco_images:
        for stream in coco_img.channels.streams():
            task = {}
            has_quantize = stream & to_quantize
            has_remove = stream & to_remove
            obj = coco_img.find_asset_obj(stream)
            old_quantization = obj.get('quantization', None)
            if has_remove.numel():
                has_quantize = has_quantize - has_remove
            if has_remove is not None and has_remove.numel():
                task['has_remove'] = has_remove
            if old_quantization is None:
                # Dont redo quantization
                if has_quantize.numel():
                    total_modified = has_quantize.as_oset() | has_remove.as_oset()
                    task['has_quantize'] = has_quantize
                    if len(total_modified) != len(stream):
                        missing = stream - total_modified
                        msg = f'Must quantize or remove entire stream: {stream=!s}, {has_quantize=!s}, {has_remove=!s}, {missing=!s}'
                        task['error'] = msg
            if task:
                task['gid'] = coco_img.img['id']
                task['stream'] = stream
                tasks.append(task)

    if len(tasks) == 0:
        print('No modifications need to be made')
    else:
        print('Found {} tasks'.format(len(tasks)))
        errors = [t['error'] for t in tasks if t.get('error', None)]
        if errors:
            for msg in errors:
                print(msg)
            raise ValueError('Cannot modify')

        gid_to_tasks = ub.group_items(tasks, lambda x: x['gid'])

        bundle_dpath = ub.Path(dset.bundle_dpath)
        jobs = ub.JobPool('process', max_workers=config['workers'])

        for gid, sub_tasks in ub.ProgIter(list(gid_to_tasks.items()), desc='submit reformat jobs'):
            coco_img = dset.coco_image(gid)
            for task in sub_tasks:
                stream = task['stream']
                has_remove = task.get('has_remove', None)
                has_quantize = task.get('has_quantize', None)
                obj = coco_img.find_asset_obj(stream)

                job = jobs.submit(reformat_obj, obj, bundle_dpath, has_remove,
                                  has_quantize)
                job.obj = obj

        finalize_tasks = []
        for job in jobs.as_completed(desc='reformat images'):
            fpath, new_fpath, new_obj = job.result()
            # Update the coco object in memory
            job.obj.update(new_obj)
            finalize_tasks.append({
                'dst': fpath,
                'src': new_fpath,
            })

        tmp_fpath = ub.Path(dset.fpath).augment(suffix='.movtmp')
        dset.dump(tmp_fpath, newlines=True)

        # Point of no return

        # Finalize everything by executing move
        os.rename(tmp_fpath, dset.fpath)
        for task in ub.ProgIter(finalize_tasks, desc='finalize image move'):
            os.rename(task['src'], task['dst'])


def reformat_obj(obj, bundle_dpath, has_remove, has_quantize):
    import xarray as xr
    from kwcoco.channel_spec import FusedChannelSpec
    import kwimage
    import numpy as np
    quantize_dtype = np.int16
    quantize_max = np.iinfo(quantize_dtype).max
    quantize_min = 0
    quantize_nan = -9999

    fname = ub.Path(obj['file_name'])
    fpath = bundle_dpath / fname

    # Write to a temporary sidecar so we dont clobber anything until
    # we know the entire operation worked
    new_fpath = fpath.augment(suffix='.modtmp')

    # Might be able to do this with a gdal command instead
    ma_imdata = kwimage.imread(fpath, nodata='ma')
    imdata = ma_imdata.data
    mask = ma_imdata.mask
    new_obj = obj.copy()

    assert obj['channels'] is not None
    obj_channels = FusedChannelSpec.coerce(obj['channels'])

    if has_remove is not None and has_remove.numel():
        keep_channels = obj_channels - has_remove
        coords = {}
        if obj_channels is not None:
            coords['c'] = obj_channels.as_list()
        xr_imdata = xr.DataArray(imdata, dims=('y', 'x', 'c'), coords=coords)
        xr_mask = xr.DataArray(mask, dims=('y', 'x', 'c'), coords=coords)
        imdata_ = xr_imdata.loc[:, :, keep_channels].data
        mask_ = xr_mask.loc[:, :, keep_channels].data
        new_obj.update({
            'channels': keep_channels.concise().spec,
            'num_bands': keep_channels.numel(),
        })
    else:
        keep_channels = obj_channels
        imdata_ = imdata
        mask_ = mask

    prev_quant = obj.get('quantization', None)
    if prev_quant is None and has_quantize is not None and has_quantize.numel():
        assert prev_quant is None
        assert has_quantize.as_set() == keep_channels.as_set()
        if imdata.dtype.kind == 'f':
            max_val = ma_imdata.max()
            min_val = ma_imdata.min()
            if min_val < 0 or max_val > 1:
                raise Exception('Can only quantize 0-1 float arrays')
            old_max = 1
            old_min = 0
        else:
            raise NotImplementedError

        old_extent = (old_max - old_min)
        new_extent = (quantize_max - quantize_min)
        quant_factor = new_extent / old_extent

        quantization = {
            'orig_min': old_min,
            'orig_max': old_max,
            'quant_min': quantize_min,
            'quant_max': quantize_max,
            'nodata': quantize_nan,
        }
        new_imdata = (imdata_ - old_min) * quant_factor + quantize_min
        new_imdata = new_imdata.astype(quantize_dtype)
        new_imdata[mask_] = quantize_nan

        # Denote that this channel is quantized
        new_obj.update({
            'quantization': quantization,
        })
    else:
        new_imdata = imdata_

    kwimage.imwrite(new_fpath, new_imdata, backend='gdal',
                    compress='DEFLATE', blocksize=128,
                    nodata=quantize_nan)

    return fpath, new_fpath, new_obj


def schedule_quantization():
    # Temporary job
    import watch
    dvc_dpath = ub.Path(str(watch.find_smart_dvc_dpath()) + '-hdd')
    dvc_dpath / 'models/fusion/'

    pred_globpat = dvc_dpath / 'models/fusion/Drop2-Aligned-TA1-2022-02-15/*/pred_*/*/pred.kwcoco.json'
    # pred_globpat = dvc_dpath / 'models/fusion/Drop2-Aligned-TA1-2022-02-15/*/*/*/pred.kwcoco.json'

    import glob
    pred_fpaths = glob.glob(str(pred_globpat))
    import cmd_queue
    queue = cmd_queue.Queue.create(backend='tmux', name='quantize-jobs', size=8)
    for pred_fpath in pred_fpaths:
        queue.submit(ub.codeblock(
            rf'''
            python -m watch.cli.coco_reformat_channels \
                --src={pred_fpath} \
                --quantize="not_salient|salient,Site Preparation|Active Construction|Post Construction|No Activity" \
                --remove="not_salient,negative,ignore,negative,positive,background|Unknown" \
                --workers=0
            '''))
    queue.rprint()
    queue.run()
    queue.monitor()


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/watch/cli/coco_combine_features.py
    """
    main(cmdline=True)



import json
import logging
import warnings
from functools import partial
import ubelt as ub

import os
import click
import kwcoco
import kwimage
import numpy as np
import torch
# import torchvision.transforms
# from medpy.filter.smoothing import anisotropic_diffusion
from torch.utils.data import DataLoader
from tqdm import tqdm

from scipy import ndimage

from . import modules_monkeypatch  # NOQA
from .datasets import WVRgbDataset, WVSuperRgbDataset
# from .pl_highres_verify import MultiTaskModel, modify_bn, dfactor, local_utils
from .pl_highres_verify import MultiTaskModel, modify_bn, dfactor
from .utils import process_image_chunked
from ..landcover.detector import get_device
from ..landcover.predict import get_output_file
from ..landcover.utils import setup_logging

log = logging.getLogger(__name__)


# Debug variables
ENABLE_WRITES = 1  # set to zero to disable writing to disk
ENABLE_MODEL = 1  # set to zero to disable running the model forward
ENABLE_PROCESS = 1  # set to zero to disable chunked process
ENABLE_CHUNKS = 1  # set to zero to disable chunking (increases VRAM)


@click.command()
@click.option('--dataset', required=True, type=click.Path(exists=True), help='input kwcoco dataset')
@click.option('--deployed', required=True, type=click.Path(exists=True), help='pytorch weights file')
@click.option('--output', required=False, type=click.Path(), help='output kwcoco dataset')
@click.option('--window_size', required=False, type=int, default=1024, help='sliding window size')
@click.option('--dump_shards', required=False, default=False, help='if True, output partial kwcoco files as they are completed')
@click.option('--data_workers', required=False, default=0, help='background data loaders')
@click.option('--select_images', required=False, default=None, help='if specified, a jq operation to filter images')
@click.option('--select_videos', required=False, default=None, help='if specified, a jq operation to filter videos')
@click.option('--asset_suffix', required=False, default='_assets/dzyne_depth', help='folder relative to output to save features in')
@click.option('--cache', required=False, default=0, help='if True, enable caching of results')
@click.option('--use_super_res_bands', required=False, default=False, help='if True, uses super-res (upsampled) bands to form source RGB image')
@click.option('--scale', required=False, default=4, help='Specify scaling factor between image size and super-res bands, only applicable when use_super_res_bands is true')
def predict(dataset, deployed, output, window_size=2048, dump_shards=False,
            data_workers=0, select_images=None, select_videos=None,
            asset_suffix='_assets/dzyne_depth', cache=False, use_super_res_bands=False, scale=4):
    """
    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from watch.tasks.depth.predict import *  # NOQA
        >>> import watch
        >>> dvc_dpath = watch.find_smart_dvc_dpath()
        >>> dataset = dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/data_vali.kwcoco.json'
        >>> output = dvc_dpath / 'Drop2-Aligned-TA1-2022-02-15/dzyne_depth_test.kwcoco.json'
        >>> deployed = dvc_dpath / "models/depth/weights_v1.pt"
        >>> data_workers = 0
        >>> cache = 0
        >>> select_images = '.name == "crop_20150123T020752Z_N37.734145E128.855484_N37.811709E128.946746_WV_0"'
        >>> select_videos = None
        >>> dump_shards = False
        >>> window_size = 1536
        >>> asset_suffix = '_assets/test_dzyne_depth'
        >>> predict.callback(dataset=dataset, deployed=deployed, output=output,
        >>>         window_size=window_size, dump_shards=dump_shards,
        >>>         data_workers=data_workers, select_images=select_images,
        >>>         select_videos=select_videos, cache=cache)
    """
    weights_filename = ub.Path(deployed)

    output_dset_filename = ub.Path(get_output_file(output))

    output_bundle_dpath = output_dset_filename.parent
    output_data_dir = output_bundle_dpath / asset_suffix

    log.info('Input:          {}'.format(dataset))
    log.info('Weights:        {}'.format(weights_filename))
    log.info('Output:         {}'.format(output_dset_filename))
    log.info('Output Images:  {}'.format(output_data_dir))

    input_dset = kwcoco.CocoDataset.coerce(dataset)
    input_bundle_dpath = ub.Path(input_dset.bundle_dpath)

    from watch.utils import kwcoco_extensions
    from watch.tasks.fusion.predict import quantize_float01
    filtered_gids = kwcoco_extensions.filter_image_ids(
        input_dset,
        include_sensors=['WV'],
        select_images=select_images,
        select_videos=select_videos
    )
    log.info('Valid Images:          {}'.format(len(filtered_gids)))
    input_dset = input_dset.subset(filtered_gids)

    output_dset = input_dset.copy()
    if input_bundle_dpath != output_bundle_dpath:
        # Need to change the root of the output directory
        # The kwcoco reroot logic is flakey for complex cases, so be careful
        # In the normal case where the output and input kwcoco share the same
        # bundle, then this logic is avoided
        output_dset.reroot(absolute=True)
        output_dset.fpath = str(output_dset_filename)
        new_prefix = os.path.relpath(input_bundle_dpath, output_bundle_dpath)
        output_dset.reroot(old_prefix=str(input_bundle_dpath),
                           new_prefix=str(new_prefix), absolute=False,
                           check=True)

    # input data
    if use_super_res_bands is True:
        torch_dataset = WVSuperRgbDataset(dset=input_dset, scale=scale)  # TODO - auto-detect scaling
    else:
        torch_dataset = WVRgbDataset(input_dset)

    if cache:
        log.debug('checking for cached files')

        # Remove any image ids that are already computed
        gid_to_pred_filename = {}
        miss_gids = []
        hit_gids = []
        for gid in torch_dataset.gids:
            img_info = torch_dataset.dset.imgs[gid]
            pred_filename = _image_pred_filename(torch_dataset,
                                                 output_data_dir, img_info)
            gid_to_pred_filename[gid] = pred_filename
            if pred_filename.exists():
                hit_gids.append(gid)
            else:
                miss_gids.append(gid)

        log.info(
            f'Found {len(hit_gids)} / {len(gid_to_pred_filename)} cached depth maps')
        # Might be a better way to indicate a subset, but this works
        torch_dataset.gids = miss_gids

    # model
    log.debug('loading model')
    config = _load_config()
    config['backbone_params']['pretrained'] = False  # dont download on predict
    model = MultiTaskModel(config=config)
    state_dict = torch.load(
        weights_filename,
        map_location=lambda storage,
        loc: storage)
    model.load_state_dict(state_dict)

    model = modify_bn(model, track_running_stats=False, bn_momentum=0.01)
    model = model.eval()
    model.to(get_device())

    S = window_size
    chip_size = (S, S, 3)
    overlap = (128, 128, 0)
    output_dtype = np.float32  # Will be quantized as a final step

    process_func = partial(run_inference, model=model)

    log.debug('processing images')
    dataloader = DataLoader(torch_dataset, num_workers=data_workers,
                            batch_size=1, pin_memory=1, collate_fn=lambda x: x)
    with torch.no_grad():
        for batch in tqdm(dataloader, miniters=1, unit='image', disable=False):
            assert len(batch) == 1
            batch_item = batch[0]
            gid = batch_item['id']

            # get clean img_info
            img_info = torch_dataset.dset.imgs[gid]

            pred_filename = _image_pred_filename(torch_dataset,
                                                 output_data_dir, img_info)
            if cache and pred_filename.exists():
                # Dereference items after we are done with them
                batch_item = None
                image = None
                continue

            try:
                image = batch_item['imgdata']
                log.info('processing image {}'.format(image.shape))

                if ENABLE_PROCESS:
                    use_chunks = ENABLE_CHUNKS
                    if image.shape[0] < S and image.shape[1] < S:
                        use_chunks = 0

                    if use_chunks:
                        pred = process_image_chunked(image, process_func,
                                                     chip_size=chip_size,
                                                     overlap=overlap,
                                                     output_dtype=output_dtype)
                    else:
                        pred = process_func(image)
                else:
                    pred = np.zeros(image.shape[0:2], dtype=output_dtype)
                # Dereference items after we are done with them
                batch_item = None  # dereference for memory
                image = None  # dereference for memory

                quant_pred, quantization = quantize_float01(
                    pred, old_min=0, old_max=1, quantize_dtype=np.uint8)
                pred = None  # dereference for memory

                info = _write_output(img_info, quant_pred, pred_filename,
                                     output_bundle_dpath, quantization)
                quant_pred = None  # dereference for memory

                aux = output_dset.imgs[gid].get('auxiliary', [])
                aux.append(info)
                output_dset.imgs[gid]['auxiliary'] = aux

                if dump_shards:
                    # Dump debugging shard
                    shard_dset = output_dset.subset([gid])
                    shard_dset.reroot(absolute=True)
                    shard_dset.fpath = pred_filename.augment(
                        ext='.kwcoco.json')
                    # output_dpath / (imgname + '_depth.kwcoco.json')
                    shard_dset.dump(shard_dset.fpath, indent=2)

            except KeyboardInterrupt:
                log.info('interrupted')
                break
            except Exception:
                log.exception(
                    'Unable to load id:{} - {}'.format(img_info['id'], img_info['name']))

    if cache and hit_gids:
        from watch.utils import util_gdal
        # add metadata for cache items
        for gid in hit_gids:
            img_info = torch_dataset.dset.imgs[gid]
            pred_filename = _image_pred_filename(torch_dataset,
                                                 output_data_dir, img_info)
            with util_gdal.GdalDataset.open(pred_filename, 'r') as gdal_img:
                pred_shape = (gdal_img.RasterYSize, gdal_img.RasterXSize,
                              gdal_img.RasterCount)
            # pred_shape = kwimage.load_image_shape(pred_filename)

            # Hack to get the quantization dict that would have been computed
            # at predict time.
            _, quantization = quantize_float01(
                None, old_min=0, old_max=1, quantize_dtype=np.uint8)

            info = _build_aux_info(img_info, pred_shape, pred_filename,
                                   output_bundle_dpath, quantization)
            aux = output_dset.imgs[gid].get('auxiliary', [])
            aux.append(info)
            output_dset.imgs[gid]['auxiliary'] = aux

    if ENABLE_WRITES:
        output_dset.dump(str(output_dset_filename), indent=2)
        output_dset.validate()
        log.info('output written to {}'.format(output_dset_filename))


def _image_pred_filename(torch_dataset, output_data_dir, img_info):
    # Construct an output file name based on the video and image name
    imgname = img_info['name']
    vidid = img_info.get('video_id', None)
    if vidid is not None:
        vidname = torch_dataset.dset.index.videos[vidid]['name']
        output_dpath = output_data_dir / vidname
    else:
        output_dpath = output_data_dir
    pred_filename = output_dpath / (imgname + '_depth.tif')
    return pred_filename


def fake_model(batch2, tta=True):
    # For testing
    np_data = batch2['image'][0].permute(1, 2, 0).numpy()
    x = kwimage.gaussian_blur(np_data, sigma=7)
    depth = torch.from_numpy(x.mean(axis=2))[None, None]
    pred2 = dict(depth=depth, seg=depth)
    return pred2, batch2


def _test():
    """
    Small test to check that stitching logic works when nan regions are
    involved.
    """
    import kwimage
    src = kwimage.ensure_float01(kwimage.grab_test_image(dsize=(2048, 2048)))
    nan_poly = kwimage.Polygon.random(rng=32021).scale(src.shape[0] * 3)
    image = nan_poly.fill(src.copy() * 255, np.nan)
    output_dtype = np.uint8
    overlap = (0, 0, 0)
    chip_size = (512, 512, 3)
    model = fake_model
    process_func = partial(run_inference, model=model, device='cpu')
    pred = process_image_chunked(image, process_func,
                                 chip_size=chip_size,
                                 overlap=overlap,
                                 output_dtype=output_dtype)

    from watch.tasks.fusion.predict import quantize_float01
    quant_pred, quantization = quantize_float01(pred, old_min=0, old_max=1,
                                                quantize_dtype=np.uint8)
    print('quantization = {}'.format(ub.urepr(quantization, nl=1)))

    import kwplot
    kwplot.autompl()
    kwplot.imshow(kwimage.normalize_intensity(
        image), pnum=(1, 3, 1), doclf=True)
    kwplot.imshow(pred, pnum=(1, 3, 2))
    kwplot.imshow(quant_pred, pnum=(1, 3, 3))


def run_inference(image, model, device=0):
    """
    Example:
        >>> from watch.tasks.depth.predict import *  # NOQA
        >>> import kwimage
        >>> import kwarray
        >>> src = kwimage.ensure_float01(kwimage.grab_test_image(dsize=(512, 512)))
        >>> src = kwimage.Polygon.random(rng=None).scale(src.shape[0]).fill(src.copy(), np.nan)
        >>> model = fake_model
        >>> image = src * 255
        >>> device = 'cpu'
        >>> result = run_inference(image, model, device=device)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> kwplot.imshow(src, pnum=(1, 2, 1), doclf=True)
        >>> kwplot.imshow(result, pnum=(1, 2, 2))
    """
    if not ENABLE_MODEL:
        pred_shape = image.shape[0:2]
        return np.full(pred_shape, fill_value=np.nan, dtype=np.float32)

    img_h, img_w = image.shape[0:2]
    pad_h = pad_w = 0
    min_w, min_h = 64, 64
    if img_h < min_h or img_w < min_w:
        pad_h = min_h - img_h
        pad_w = min_w - img_w
        pad_dims = ((0, pad_h), (0, pad_w), (0, 0))
        image = np.pad(image, pad_dims, mode='constant',
                       constant_values=np.nan)

    with torch.no_grad():
        nodata_mask = np.isnan(image)
        if not np.all(nodata_mask):

            # Replace nans with zeros before going into the network
            image_float = image / 255.0  # not sure why we want to do this...

            # image_float = image.copy()
            image_float[nodata_mask] = 0
            image_tensor = torch.from_numpy(
                image_float.transpose(
                    (2, 0, 1))).contiguous()

            mean = np.nanmean(image.reshape(-1, image.shape[-1]), axis=0)
            std = np.nanstd(image.reshape(-1, image.shape[-1]), axis=0)

            batch2 = {
                "image": image_tensor[None, ...].to(device),
                "image_mean": torch.from_numpy(mean)[None, ...].to(device),
                "image_std": torch.from_numpy(std)[None, ...].to(device),
            }

            pred2, batch2 = model(batch2, tta=True)

            output_depth = pred2['depth'][0, 0, :, :].cpu().data.numpy()
            output_label = pred2['seg'][0, 0, :, :].cpu().data.numpy()
            # output_depth[nodata_mask.all(axis=2)] = np.nan

            weighted_depth = dfactor * output_depth

            alpha = 0.9
            weighted_seg = alpha * output_label + \
                (1.0 - alpha) * np.minimum(0.99, weighted_depth / 70.0)

            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                # tmp2 = 255 * anisotropic_diffusion(weighted_seg, niter=1, kappa=100, gamma=0.8)
                tmp2 = anisotropic_diffusion(
                    weighted_seg, niter=1, kappa=100, gamma=0.8)

            # weighted_final = ndimage.median_filter(tmp2.astype(np.uint8), size=7)
            weighted_final = ndimage.median_filter(tmp2, size=7)
            weighted_final[nodata_mask.all(axis=2)] = np.nan
            # weighted_final = ndimage.median_filter(tmp2.astype(np.uint8), size=7)

            if pad_h or pad_w:
                # Undo pad
                weighted_final = weighted_final[:-pad_h, :-pad_w]

        else:
            pred_shape = image.shape[0:2]
            return np.full(pred_shape, fill_value=np.nan, dtype=np.float32)

    return weighted_final


# Vendored in to deal with 3.10 issue
def anisotropic_diffusion(img, niter=1, kappa=50,
                          gamma=0.1, voxelspacing=None, option=1):
    r"""
    Edge-preserving, XD Anisotropic diffusion.


    Parameters
    ----------
    img : array_like
        Input image (will be cast to numpy.float).
    niter : integer
        Number of iterations.
    kappa : integer
        Conduction coefficient, e.g. 20-100. ``kappa`` controls conduction
        as a function of the gradient. If ``kappa`` is low small intensity
        gradients are able to block conduction and hence diffusion across
        steep edges. A large value reduces the influence of intensity gradients
        on conduction.
    gamma : float
        Controls the speed of diffusion. Pick a value :math:`<= .25` for stability.
    voxelspacing : tuple of floats or array_like
        The distance between adjacent pixels in all img.ndim directions
    option : {1, 2, 3}
        Whether to use the Perona Malik diffusion equation No. 1 or No. 2,
        or Tukey's biweight function.
        Equation 1 favours high contrast edges over low contrast ones, while
        equation 2 favours wide regions over smaller ones. See [1]_ for details.
        Equation 3 preserves sharper boundaries than previous formulations and
        improves the automatic stopping of the diffusion. See [2]_ for details.

    Returns
    -------
    anisotropic_diffusion : ndarray
        Diffused image.

    Notes
    -----
    Original MATLAB code by Peter Kovesi,
    School of Computer Science & Software Engineering,
    The University of Western Australia,
    pk @ csse uwa edu au,
    <http://www.csse.uwa.edu.au>

    Translated to Python and optimised by Alistair Muldal,
    Department of Pharmacology,
    University of Oxford,
    <alistair.muldal@pharm.ox.ac.uk>

    Adapted to arbitrary dimensionality and added to the MedPy library Oskar Maier,
    Institute for Medical Informatics,
    Universitaet Luebeck,
    <oskar.maier@googlemail.com>

    June 2000  original version. -
    March 2002 corrected diffusion eqn No 2. -
    July 2012 translated to Python -
    August 2013 incorporated into MedPy, arbitrary dimensionality -

    References
    ----------
    .. [1] P. Perona and J. Malik.
       Scale-space and edge detection using ansotropic diffusion.
       IEEE Transactions on Pattern Analysis and Machine Intelligence,
       12(7):629-639, July 1990.
    .. [2] M.J. Black, G. Sapiro, D. Marimont, D. Heeger
       Robust anisotropic diffusion.
       IEEE Transactions on Image Processing,
       7(3):421-432, March 1998.
    """
    # define conduction gradients functions
    import numpy
    if option == 1:
        def condgradient(delta, spacing):
            return numpy.exp(-(delta / kappa)**2.) / float(spacing)
    elif option == 2:
        def condgradient(delta, spacing):
            return 1. / (1. + (delta / kappa)**2.) / float(spacing)
    elif option == 3:
        kappa_s = kappa * (2**0.5)

        def condgradient(delta, spacing):
            top = 0.5 * ((1. - (delta / kappa_s)**2.)**2.) / float(spacing)
            return numpy.where(numpy.abs(delta) <= kappa_s, top, 0)

    # initialize output array
    out = numpy.array(img, dtype=numpy.float32, copy=True)

    # set default voxel spacing if not supplied
    if voxelspacing is None:
        voxelspacing = tuple([1.] * img.ndim)

    # initialize some internal variables
    deltas = [numpy.zeros_like(out) for _ in range(out.ndim)]

    for _ in range(niter):

        # calculate the diffs
        for i in range(out.ndim):
            slicer = tuple([slice(None, -1) if j == i else slice(None)
                           for j in range(out.ndim)])
            deltas[i][slicer] = numpy.diff(out, axis=i)

        # update matrices
        matrices = [
            condgradient(
                delta,
                spacing) *
            delta for delta,
            spacing in zip(
                deltas,
                voxelspacing)]

        # subtract a copy that has been shifted ('Up/North/West' in 3D case) by one
        # pixel. Don't as questions. just do it. trust me.
        for i in range(out.ndim):
            slicer = tuple([slice(1, None) if j == i else slice(None)
                           for j in range(out.ndim)])
            matrices[i][slicer] = numpy.diff(matrices[i], axis=i)

        # update the image
        out += gamma * (numpy.sum(matrices, axis=0))

    return out


def _build_aux_info(img_info, pred_shape, pred_filename, output_bundle_dpath,
                    quantization):
    info = {
        'file_name': str(pred_filename.relative_to(output_bundle_dpath)),
        'channels': 'depth',
        'height': pred_shape[0],
        'width': pred_shape[1],
        'num_bands': 1,
        'quantization': quantization,
        'warp_aux_to_img': {'scale': [img_info['width'] / pred_shape[1],
                                      img_info['height'] / pred_shape[0]],
                            'type': 'affine'}
    }
    return info


def _write_output(img_info, pred, pred_filename,
                  output_bundle_dpath, quantization):
    pred_shape = pred.shape
    info = _build_aux_info(img_info, pred_shape, pred_filename,
                           output_bundle_dpath, quantization)

    # with warnings.catch_warnings():
    #     warnings.simplefilter('ignore', UserWarning)
    if ENABLE_WRITES:
        pred_filename.parent.mkdir(parents=True, exist_ok=True)

        kwimage.imwrite(str(pred_filename),
                        pred, backend='gdal', blocksize=256,
                        nodata=quantization['nodata'],
                        compress='DEFLATE', overviews=3)
    return info


def _load_config():
    from importlib import resources as importlib_resources
    fp = importlib_resources.open_text('watch.tasks.depth', 'config.json')
    return json.load(fp)


if __name__ == '__main__':
    r"""
    # Notes:

        weights_v1 - for RGB
        weights_v2_gray - for PAN

        TODO: Predict with both models, one for RGB and one with PAN

    # VRAM usage with weights_v2_gray
    # window_size=512:   4.951 GB
    # window_size=640:   7.406 GB
    # window_size=704:   8.912 GB
    # window_size=736:   9.310 GB
    # window_size=768:  10.099 GB
    # window_size=1024: 17.111 GB
    # window_size=1152: 21.007 GB

    ## Drop 4 ##
    python3 -m watch.tasks.depth.predict \
    --deployed=/smart/backup/models/depth/weights_v1.pt \
    --dataset=/output/Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC/data_wv_superRes.kwcoco.json \
    --output=/output/Aligned-Drop4-2022-08-08-TA1-S2-WV-PD-ACC/data_wv_superRes_depth.kwcoco.json \
    --data_workers=8 \
    --window_size=2048 \
    --cache=1 \
    --use_super_res_bands=True \
    --scale=4

    python -m watch visualize $KWCOCO_BUNDLE_DPATH/dzyne_depth.kwcoco.json \
        --animate=True --channels="depth,red|green|blue" --skip_missing=True \
        --select_images '.sensor_coarse == "WV"' --workers=4 --draw_anns=False

    python -m watch stats $KWCOCO_BUNDLE_DPATH/dzyne_depth.kwcoco.json

    python -m kwcoco stats $KWCOCO_BUNDLE_DPATH/dzyne_depth.kwcoco.json


    Notes:
        Does export MALLOC_MMAP_THRESHOLD_=0 help with memory?

        export MALLOC_MMAP_THRESHOLD_=8192
        export MALLOC_ARENA_MAX=4

    """
    setup_logging()
    torch.hub.set_dir('/tmp/weights')
    predict()



import logging


def setup_logging(level=logging.WARNING):
    logging.basicConfig(
        level=level,
        format="%(asctime)s - %(name)-40s - %(levelname)-7s - %(message)s",
        handlers=[logging.StreamHandler()]
    )
    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    logging.getLogger('rasterio').setLevel(logging.INFO)



"""
Defines a torch Dataset and lightning DataModule for kwcoco video data.

The parameters to each are handled by scriptconfig objects, which prevents us
from needing to specify what the available options are in multiple places.
"""
from watch.tasks.fusion.datamodules.kwcoco_dataset import *  # NOQA
from watch.tasks.fusion.datamodules.kwcoco_datamodule import *  # NOQA



import torch.nn as nn

from watch.tasks.rutgers_material_change_detection.models.apnb import AFNB


class Attention(nn.Module):
    def __init__(self, framework_name, feat_sizes, active_layers, n_heads):
        super(Attention, self).__init__()
        self.n_heads = n_heads
        self.feat_sizes = feat_sizes
        self.active_layers = active_layers
        self.framework_name = framework_name

        self.build()

    def build(self):
        raise NotImplementedError("Build method for attention not implemented.")

    def forward(self):
        pass


class SelfAttention(Attention):
    def __init__(self, framework_name, feat_sizes, active_layers, n_heads):
        super(SelfAttention, self).__init__(framework_name, feat_sizes, active_layers, n_heads)

    def build(self):
        # Generate Attention per frame.
        self.attention_layers = nn.ModuleDict()
        for i, feat_size in enumerate(self.feat_sizes):
            self.attention_layers["layer" + str(i + 1)] = nn.TransformerEncoderLayer(
                feat_size[0], self.n_heads, dim_feedforward=feat_size[0] * 2
            )

    def forward(self, feats):
        refined_feats = {}
        for layer_name, feat in feats.items():
            batch_size, feat_size, height, width = feat.shape
            feat = feat.flatten(2)  # [batch_size, feat_size, seq]
            feat = feat.permute(2, 0, 1)  # [seq, batch_size, feat_size]
            feat = self.attention_layers[layer_name](feat)  # [seq, batch_size, feat_size]
            feat = feat.permute(1, 2, 0)  # [batch_size, feat_size, seq]
            feat = feat.reshape(batch_size, feat_size, height, width)
            refined_feats[layer_name] = feat
        return refined_feats


class AsymmetricPyramidSelfAttention(Attention):
    def __init__(self, framework_name, feat_sizes, active_layers, n_heads):
        super(AsymmetricPyramidSelfAttention, self).__init__(framework_name, feat_sizes, active_layers, n_heads)

    def build(self):
        self.attention_layers = nn.ModuleDict()
        for i, feat_size in enumerate(self.feat_sizes):
            if i == 0:
                self.attention_layers["layer" + str(i + 1)] = nn.TransformerEncoderLayer(
                    feat_size[0], self.n_heads, dim_feedforward=feat_size[0] * 2
                )
            else:
                prev_channels = self.feat_sizes[i - 1][0]
                curr_channels = self.feat_sizes[i][0]
                self.attention_layers["layer" + str(i + 1)] = AFNB(
                    low_in_channels=prev_channels,
                    high_in_channels=curr_channels,
                    out_channels=curr_channels,
                    key_channels=prev_channels,
                    value_channels=prev_channels,
                    dropout=0.0,
                    norm_type="batchnorm",
                )

    def forward(self, feats):
        refined_feats = {}
        for i, (layer_name, feat) in enumerate(feats.items()):
            if i == 0:
                batch_size, feat_size, height, width = feat.shape
                feat = feat.flatten(2)  # [batch_size, feat_size, seq]
                feat = feat.permute(2, 0, 1)  # [seq, batch_size, feat_size]
                feat = self.attention_layers[layer_name](feat)  # [seq, batch_size, feat_size]
                feat = feat.permute(1, 2, 0)  # [batch_size, feat_size, seq]
                feat = feat.reshape(batch_size, feat_size, height, width)
                refined_feats[layer_name] = feat
            else:
                prev_feat = feats["layer" + str(i)]
                curr_feat = feat
                refined_feats[layer_name] = self.attention_layers[layer_name](prev_feat, curr_feat)
        return refined_feats



from collections import defaultdict

import torch
import numpy as np

from watch.tasks.rutgers_material_change_detection.models.base_model import BaseFramework


class LateFusionFramework(BaseFramework):
    def __init__(self, task_mode, encoder, decoder, attention, sequence_model, **kwargs):
        super(LateFusionFramework, self).__init__(task_mode, encoder, decoder, attention)

        self.sequence_model = sequence_model
        cfg = kwargs["cfg"]

        # Handle material reconstruction discritzation.
        if task_mode == "ss_mat_recon":
            if cfg.loss.discretize:
                k_classes = cfg.loss.k_classes
                self.nn_totems, self.color_codes = {}, {}
                for i, feat_size in enumerate(kwargs["feat_sizes"]):
                    feat_dim = feat_size[0]
                    self.nn_totems["layer" + str(i + 1)] = torch.randn(size=[feat_dim, k_classes])
                    self.nn_totems["layer" + str(i + 1)] = (
                        self.nn_totems["layer" + str(i + 1)]
                        / torch.norm(self.nn_totems["layer" + str(i + 1)], p=2, dim=0)
                    ).to(kwargs["device"])
                    self.color_codes["layer" + str(i + 1)] = np.random.rand(k_classes, 3)
            else:
                self.nn_totems = None
        else:
            self.nn_totems = None

        # Feature differencing option.
        self.feature_difference = cfg.framework.feat_difference

    def forward(self, data):
        """Forward pass.

        Args:
            data (dict): [description]

        Returns:
            [type]: [description]
        """
        output = {}

        video = data["video"]  # [batch_size, frames, channels, height, width]

        # Separate frames of video.
        batch_size, n_frames, n_channels, height, width = video.shape
        frames = [video[:, i] for i in range(n_frames)]

        # Pass frames through encoder separately.
        feats = []
        for frame in frames:
            feat = self.encoder(frame)
            feats.append(feat)

        # Stack features per layer.
        stacked_feats = defaultdict(list)
        for feat_dict in feats:
            for key, value in feat_dict.items():
                stacked_feats[key].append(value)

        # If TRUE, compute feature difference.
        if self.feature_difference:
            for layer_name, feat_list in stacked_feats.items():
                compare_feat = feat_list[0]
                diff_feats = []
                for feat in feat_list:
                    diff_feat = feat - compare_feat
                    diff_feats.append(diff_feat)
                stacked_feats[layer_name] = diff_feats

        # TODO: Add option to add relative positional information here.

        # Pass encoder features through sequence model.
        if self.sequence_model == "mean_feats":
            ## Concatenate and compute mean features.
            for key, value in stacked_feats.items():
                stacked_feats[key] = torch.mean(torch.stack(stacked_feats[key], dim=0), dim=0)

            if self.attention:
                stacked_feats = self.attention(stacked_feats)

        elif self.sequence_model == "max_feats":
            # Concatenate and compute max features.
            for key, value in stacked_feats.items():
                stacked_feats[key] = torch.max(torch.stack(stacked_feats[key], dim=0), dim=0)[0]

            if self.attention:
                stacked_feats = self.attention(stacked_feats)

        elif self.sequence_model is None:
            pass

        elif type(self.sequence_model) is not str:
            # Concatenate and compute features.
            for layer_name, seq_name in zip(stacked_feats.keys(), self.sequence_model.keys()):
                # Pass fetures into sequence model.
                if type(self.sequence_model[seq_name]) is str:
                    stacked_feats[layer_name] = torch.max(torch.stack(stacked_feats[layer_name], dim=0), dim=0)[0]
                else:
                    seq_output = self.sequence_model[seq_name](torch.stack(stacked_feats[layer_name], dim=1))

                    # Reassemble the output tokens into OG feature shape.
                    device = stacked_feats[layer_name][0].device
                    n_out_channels = stacked_feats[layer_name][0].shape[1]
                    canvas = torch.zeros([batch_size, n_out_channels, height, width]).to(device)

                    for feat_token in seq_output:
                        # Fill prediction into output canvas.
                        H0, W0, dH, dW = feat_token["crop_slice"]
                        canvas[:, :, H0 : H0 + dH, W0 : W0 + dW] = feat_token["prediction"][:, -1].reshape(
                            [batch_size, n_out_channels, dH, dW]
                        )
                        del feat_token

                    stacked_feats[layer_name] = canvas
        else:
            raise NotImplementedError(
                f'Sequence model name "{self.sequence_model}" is not implemented for Late Fusion.'
            )

        # Discretize features.
        if self.nn_totems is not None:
            material_cluster_ids = {}
            for layer_name, layer_feats in stacked_feats.items():
                if type(layer_feats) is list:
                    totem_layer_feats, layer_cluster_ids = [], []
                    for layer_feat in layer_feats:
                        norm_values = torch.norm(layer_feat, p=2, dim=1)
                        layer_feat = layer_feat / norm_values.unsqueeze(1)
                        batch_size, n_channels, height, width = layer_feat.shape

                        layer_feat = layer_feat.permute(0, 2, 3, 1)
                        nn_totem = self.nn_totems[layer_name]

                        feat_class_sim = torch.matmul(layer_feat, nn_totem)
                        _, cluster_ids = torch.max(feat_class_sim, dim=-1)

                        layer_cluster_ids.append(feat_class_sim)

                        totem_feats = (
                            torch.index_select(nn_totem, 1, cluster_ids.flatten())
                            .reshape([n_channels, batch_size, height, width])
                            .permute(1, 0, 2, 3)
                        )
                        totem_layer_feats.append(totem_feats)
                    material_cluster_ids[layer_name] = torch.stack(layer_cluster_ids, dim=0)
                else:
                    breakpoint()
                    pass
                pass
            output["mat_cluster_ids"] = material_cluster_ids

        if self.task_mode == "ss_mat_recon":
            # Reorganize features.
            n_frames = len(stacked_feats["layer1"])
            all_frame_feats = []
            for i in range(n_frames):
                frame_feats = {}
                for layer_name in stacked_feats.keys():
                    frame_feats[layer_name] = stacked_feats[layer_name][i]
                all_frame_feats.append(frame_feats)

            # Separately recompute each frame of the video.
            if self.decoder.name == "unet":
                frame_preds = []
                for i, frame_feat in enumerate(all_frame_feats):
                    frame_pred = self.decoder(frame_feat, video[:, i])
                    frame_preds.append(frame_pred)
                pred = torch.stack(frame_preds, dim=1)
            else:
                frame_preds = []
                for i, frame_feat in enumerate(all_frame_feats):
                    frame_pred = self.decoder(frame_feat)
                    frame_preds.append(frame_pred)
                pred = torch.stack(frame_preds, dim=1)
        else:
            # Use fused features to generate prediction.
            if self.decoder.name == "unet":
                pred = self.decoder(stacked_feats, video[:, -1])
            else:
                pred = self.decoder(stacked_feats)

        output[self.task_mode] = pred
        return output



import torch
from torch import nn

from einops import rearrange, repeat
from einops.layers.torch import Rearrange

# helpers


def pair(t):
    return t if isinstance(t, tuple) else (t, t)


# classes


class PreNorm(nn.Module):
    def __init__(self, dim, fn):
        super().__init__()
        self.norm = nn.LayerNorm(dim)
        self.fn = fn

    def forward(self, x, **kwargs):
        return self.fn(self.norm(x), **kwargs)


class FeedForward(nn.Module):
    def __init__(self, dim, hidden_dim, dropout=0.0):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, hidden_dim), nn.GELU(), nn.Dropout(dropout), nn.Linear(hidden_dim, dim), nn.Dropout(dropout)
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, dropout=0.0):
        super().__init__()
        inner_dim = dim_head * heads
        project_out = not (heads == 1 and dim_head == dim)

        self.heads = heads
        self.scale = dim_head**-0.5

        self.attend = nn.Softmax(dim=-1)
        self.to_qkv = nn.Linear(dim, inner_dim * 3, bias=False)

        self.to_out = nn.Sequential(nn.Linear(inner_dim, dim), nn.Dropout(dropout)) if project_out else nn.Identity()

    def forward(self, x):
        qkv = self.to_qkv(x).chunk(3, dim=-1)
        q, k, v = map(lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads), qkv)

        dots = torch.matmul(q, k.transpose(-1, -2)) * self.scale

        attn = self.attend(dots)

        out = torch.matmul(attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)


class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout=0.0):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        PreNorm(dim, Attention(dim, heads=heads, dim_head=dim_head, dropout=dropout)),
                        PreNorm(dim, FeedForward(dim, mlp_dim, dropout=dropout)),
                    ]
                )
            )

    def forward(self, x):
        for attn, ff in self.layers:
            x = attn(x) + x
            x = ff(x) + x
        return x


class ViT(nn.Module):
    def __init__(
        self,
        *,
        image_size,
        patch_size,
        num_classes,
        dim,
        depth,
        heads,
        mlp_dim,
        pool="cls",
        channels=3,
        dim_head=64,
        dropout=0.0,
        emb_dropout=0.0
    ):
        super().__init__()
        image_height, image_width = pair(image_size)
        patch_height, patch_width = pair(patch_size)

        assert (
            image_height % patch_height == 0 and image_width % patch_width == 0
        ), "Image dimensions must be divisible by the patch size."

        num_patches = (image_height // patch_height) * (image_width // patch_width)
        patch_dim = channels * patch_height * patch_width
        assert pool in {"cls", "mean"}, "pool type must be either cls (cls token) or mean (mean pooling)"

        self.to_patch_embedding = nn.Sequential(
            Rearrange("b c (h p1) (w p2) -> b (h w) (p1 p2 c)", p1=patch_height, p2=patch_width),
            nn.Linear(patch_dim, dim),
        )

        self.linear_proj_layer = nn.Linear(patch_dim, dim)

        # TODO: Update pos_embedding.
        self.pos_embedding = nn.Parameter(torch.randn(1, num_patches + 1, dim))
        self.cls_token = nn.Parameter(torch.randn(1, 1, dim))
        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(dim, depth, heads, dim_head, mlp_dim, dropout)

        self.pool = pool
        self.to_latent = nn.Identity()

        # self.mlp_head = nn.Sequential(nn.LayerNorm(dim), nn.Linear(dim, num_classes))

    def forward(self, img):
        x = self.to_patch_embedding(img)
        b, n, _ = x.shape

        cls_tokens = repeat(self.cls_token, "() n d -> b n d", b=b)
        x = torch.cat((cls_tokens, x), dim=1)
        x += self.pos_embedding[:, : (n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        x = x.mean(dim=1) if self.pool == "mean" else x[:, 0]

        x = self.to_latent(x)
        return self.mlp_head(x)

    def patch_forward(self, patches):
        """Compute the forward pass of ViT with patches instead of single frame.

        Args:
            patches (torch.Tensor): A tensor of shape [batch_size, n_patches, token_length].

        Returns:
            torch.tensor: A tensor of shape [batch_size, ]
        """
        b, n, _ = patches.shape

        patches = self.linear_proj_layer(patches)

        cls_tokens = repeat(self.cls_token, "() n d -> b n d", b=b)
        x = torch.cat((cls_tokens, patches), dim=1)
        # x += self.pos_embedding[:, :(n + 1)]
        x = self.dropout(x)

        x = self.transformer(x)

        return x




def test_small_predict_region():
    import pytest
    pytest.skip()
    from watch.tasks.fusion.fit import fit_model  # NOQA
    import ubelt as ub
    from watch.tasks.fusion.predict import predict
    import watch

    gpus = None
    test_dpath = ub.Path.appdir('watch/test/fusion/')
    results_path = (test_dpath / 'predict').ensuredir()
    ub.delete(results_path)
    ub.ensuredir(results_path)
    package_fpath = test_dpath / 'my_test_package.pt'

    train_dset = watch.coerce_kwcoco(
        'watch-msi', num_videos=3, num_frames=5, image_size=(128, 128),
        rng=3213,
    )

    # Predict via that model on a smaller dataset
    test_dset = watch.coerce_kwcoco(
        'watch-msi', num_videos=1, num_frames=1, image_size=(32, 32),
        rng=90312,
    )
    from watch.utils import kwcoco_extensions  # NOQA
    test_chans = kwcoco_extensions.coco_channel_stats(test_dset)
    train_chans = kwcoco_extensions.coco_channel_stats(train_dset)
    print('test_chans = {}'.format(ub.urepr(test_chans, nl=1)))
    print('train_chans = {}'.format(ub.urepr(train_chans, nl=1)))

    fit_kwargs = {
        'train_dataset': train_dset.fpath,
        'datamodule': 'KWCocoVideoDataModule',
        'workdir': ub.ensuredir((test_dpath, 'train')),
        'package_fpath': package_fpath,
        'max_epochs': 1,
        'time_steps': 3,
        'chip_size': 64,
        'time_sampling': 'hardish3',
        'global_change_weight': 0.0,
        'global_class_weight': 0.0,
        'global_saliency_weight': 1.0,
        'max_steps': 1,
        'learning_rate': 1e-5,
        'num_workers': 0,
        'gpus': gpus,
    }

    package_fpath = fit_model(**fit_kwargs)

    from watch.cli import torch_model_stats
    torch_model_stats.main(src=package_fpath)

    predict_kwargs = {
        'package_fpath': package_fpath,
        'pred_dataset': results_path / 'pred.kwcoco.json',
        'test_dataset': test_dset.fpath,
        'datamodule': 'KWCocoVideoDataModule',
        'write_probs': True,
        'write_preds': False,
        'with_change': False,
        'with_class': False,
        'with_saliency': True,
        'batch_size': 8,
        'tta_time': 1,
        'tta_fliprot': 0,
        'chip_overlap': 0.3,
        'num_workers': 0,
        'gpus': gpus,
    }
    result_dataset = predict(**predict_kwargs)
    pred_chans = kwcoco_extensions.coco_channel_stats(result_dataset)
    print('pred_chans = {}'.format(ub.urepr(pred_chans, nl=1)))

    from watch.cli import watch_coco_stats
    watch_coco_stats.WatchCocoStats.main(src=train_dset.fpath)
    watch_coco_stats.WatchCocoStats.main(src=train_dset.fpath)
    watch_coco_stats.WatchCocoStats.main(src=result_dataset.fpath)

    dset = result_dataset
    # Check that the result format looks correct
    for vidid in dset.index.videos.keys():
        coco_imgs = dset.images(dset.index.vidid_to_gids[1]).coco_images
        for coco_img in coco_imgs:
            assert 'salient' in coco_img.channels

    coco_img = dset.images().coco_images[0]
    # Test that new quantization does not existing APIs
    pred1 = coco_img.imdelay('salient').finalize(nodata='float')
    pred2 = coco_img.imdelay('salient').finalize(nodata='float', dequantize=False)
    assert pred1.max() <= 1
    assert pred2.max() > 1


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/tests/test_predict_small_region.py
    """
    test_small_predict_region()



from watch.monkey import monkey_lightning
from watch.monkey import monkey_tensorflow
monkey_tensorflow.disable_tensorflow_warnings()
monkey_lightning.disable_lightning_hardware_warnings()


def test_save_channelwise_plain():
    # Use one of our fusion.architectures in a test
    from watch.tasks.fusion import methods
    model = methods.MultimodalTransformer(
        arch_name="smt_it_joint_p2", input_sensorchan=5,
        change_head_hidden=0, saliency_head_hidden=0,
        class_head_hidden=0)
    _save_package(model)


def test_save_channelwise_with_dataloader():
    from watch.tasks.fusion import datamodules
    from watch.tasks.fusion import methods
    datamodule = datamodules.kwcoco_video_data.KWCocoVideoDataModule(
        train_dataset='special:vidshapes8-multispectral-multisensor', chip_size=32,
        batch_size=1, time_steps=2, num_workers=2, normalize_inputs=10, channels='auto')
    datamodule.setup('fit')
    dataset_stats = datamodule.torch_datasets['train'].cached_dataset_stats(num=3)
    classes = datamodule.torch_datasets['train'].classes

    # Use one of our fusion.architectures in a test
    model = methods.MultimodalTransformer(
        arch_name="smt_it_joint_p2", classes=classes,
        dataset_stats=dataset_stats, input_sensorchan=datamodule.input_sensorchan,
        learning_rate=1e-8, optimizer='sgd',
        change_head_hidden=0, saliency_head_hidden=0,
        class_head_hidden=0)
    _save_package_with_trainer(model, datamodule)


def test_save_heterogeneous_plain():
    # Use one of our fusion.architectures in a test
    from watch.tasks.fusion import methods
    from watch.tasks.fusion.architectures.transformer import TransformerEncoderDecoder
    position_encoder = methods.heterogeneous.ScaleAgnostictPositionalEncoder(3)
    backbone = TransformerEncoderDecoder(
        encoder_depth=1,
        decoder_depth=1,
        dim=position_encoder.output_dim + 16,
        queries_dim=position_encoder.output_dim,
        logits_dim=16,
        cross_heads=1,
        latent_heads=1,
        cross_dim_head=1,
        latent_dim_head=1,
    )
    model = methods.HeterogeneousModel(
        position_encoder=position_encoder,
        input_sensorchan=5,
        decoder="upsample",
        backbone=backbone,
    )
    _save_package(model)


def test_save_heterogeneous_with_dataloader():
    # Use one of our fusVon.architectures in a test
    from watch.tasks.fusion import methods
    from watch.tasks.fusion.architectures.transformer import TransformerEncoderDecoder
    position_encoder = methods.heterogeneous.ScaleAgnostictPositionalEncoder(3)
    backbone = TransformerEncoderDecoder(
        encoder_depth=1,
        decoder_depth=1,
        dim=position_encoder.output_dim + 16,
        queries_dim=position_encoder.output_dim,
        logits_dim=16,
        cross_heads=1,
        latent_heads=1,
        cross_dim_head=1,
        latent_dim_head=1,
    )
    model = methods.HeterogeneousModel(
        position_encoder=position_encoder,
        input_sensorchan='*:r|g|b',
        decoder="upsample",
        backbone=backbone,
    )
    from watch.tasks.fusion import datamodules
    datamodule = datamodules.kwcoco_video_data.KWCocoVideoDataModule(
        train_dataset='special:vidshapes8', chip_size=32,
        batch_size=1, time_steps=2, num_workers=2, normalize_inputs=10, channels='auto')
    datamodule.setup('fit')
    _save_package_with_trainer(model, datamodule)


def test_save_unet_plain():
    # Use one of our fusion.architectures in a test
    from watch.tasks.fusion import methods
    model = methods.UNetBaseline(
        input_sensorchan=5,
    )
    _save_package(model)


def test_save_unet_with_dataloader():
    import pytest
    pytest.skip('not working')
    # Use one of our fusion.architectures in a test
    from watch.tasks.fusion import methods
    model = methods.UNetBaseline(
        input_sensorchan='*:r|g|b',
    )
    from watch.tasks.fusion import datamodules
    datamodule = datamodules.kwcoco_video_data.KWCocoVideoDataModule(
        train_dataset='special:vidshapes8', chip_size=32,
        batch_size=1, time_steps=2, num_workers=2, normalize_inputs=10, channels='auto')
    datamodule.setup('fit')
    _save_package_with_trainer(model, datamodule)


def test_save_noop_plain():
    # Use one of our fusion.architectures in a test
    from watch.tasks.fusion import methods
    model = methods.NoopModel(
        input_sensorchan=5,)
    _save_package(model)


def test_save_noop_with_dataloader():
    # Use one of our fusion.architectures in a test
    from watch.tasks.fusion import methods
    from watch.tasks.fusion import datamodules
    model = methods.NoopModel(
        input_sensorchan=5,)
    datamodule = datamodules.kwcoco_video_data.KWCocoVideoDataModule(
        train_dataset='special:vidshapes8-multispectral-multisensor', chip_size=32,
        batch_size=1, time_steps=2, num_workers=0, normalize_inputs=10, channels='auto')
    datamodule.setup('fit')
    _save_package_with_trainer(model, datamodule)


def _save_package_with_trainer(model, datamodule):
    # Test with datamodule / trainer
    import ubelt as ub
    import pytorch_lightning as pl

    # We have to run an input through the module because it is lazy
    batch = ub.peek(iter(datamodule.train_dataloader()))
    model.training_step(batch)

    trainer = pl.Trainer(max_steps=0, accelerator='cpu', devices=1)
    trainer.fit(model=model, datamodule=datamodule)
    _save_package(model)


def _save_package(model):
    """
    Check that all of our models can be saved / reloaded via torch package.
    """
    # Test without datamodule
    import ubelt as ub
    from os.path import join
    #from watch.tasks.fusion.methods.heterogeneous import *  # NOQA
    name = model.__class__.__name__
    pkgid = id(model)
    dpath = ub.Path.appdir(f'watch/tests/package/{name}').ensuredir()
    package_path = join(dpath, f'my_package_{pkgid}.pt')

    # Save the model (TODO: need to save datamodule as well)
    model.save_package(package_path)

    # Test that the package can be reloaded
    #recon = methods.HeterogeneousModel.load_package(package_path)
    from watch.tasks.fusion.utils import load_model_from_package
    recon = load_model_from_package(package_path)
    # Check consistency and data is actually different
    recon_state = recon.state_dict()
    model_state = model.state_dict()
    assert recon is not model
    assert set(recon_state) == set(recon_state)
    from watch.utils.util_kwarray import torch_array_equal
    for key in recon_state.keys():
        assert torch_array_equal(model_state[key], recon_state[key], equal_nan=True)
        assert model_state[key] is not recon_state[key]
